#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <string.h>
#include <unistd.h>
#include "WinObj.h"

#define INIT_MENU_SIZE 64
#define GAP 8

WinObj* create_MCPopupMenu(Display* disp,
			   const char* title,
			   int bd_width,
			   Pixel fg,
			   Pixel bg,
			   CharFont* font)
{
    WinObj* menu = create_PopupMenuBase(disp, title, bd_width, fg, bg, font);
    if(menu == NULL)
	return NULL;
    menu->any.type = MCPopupMenu_T;
    SetEventMask(menu, PointerMotionMask | ButtonPressMask |
		 ButtonReleaseMask | KeyPressMask);
    return menu;
}

Bool add_MCPopupMenu(WinObj* menu,
		     const char* name,
		     int menu_id,
		     void (*selected_proc)
		     		(PopupMenuBase* menu, char* name, int id))
{
    return add_PopupMenuBase(menu, name, menu_id, selected_proc);
}

void clear_MCPopupMenu(WinObj* menu)
{
    clear_PopupMenuBase(menu);
}

Bool rename_MCPopupMenu(WinObj* win_obj, const char* name, int id)
{
    return rename_PopupMenuBase(win_obj, name, id);
}

static void calc_win_size(int n, int iw, int ih, int nline,
			  int* win_width_return,
			  int* win_height_return)
{
    *win_width_return = ((n - 1) / nline + 1) * iw;

    if(n < nline)
	*win_height_return = n * ih;
    else
	*win_height_return = nline * ih;
}

static void move_popup_window(MCPopupMenu* menu, int w, int h)
{
    int x, y;
    Display* disp = menu->disp;
    int scr = DefaultScreen(disp);

    QueryPointer(disp, menu->win, &x, &y);

    x -= w / 2;
    y -= menu->title_offset/2 + 5;

    if(x < 0)
	x = 0;
    if(y < 0)
	y = 0;

    else if(x + w >= DisplayWidth(disp, scr))
	x = DisplayWidth(disp, scr) - w - 1;

    if(y + h >= DisplayHeight(disp, scr))
	y = DisplayHeight(disp, scr) - h - 1;
    MoveWinObj((WinObj *)menu, x, y - 5);
}

static void draw_menu_title(MCPopupMenu* menu, int ih)
{
    int bd = menu->bd_width;
    DrawRuledHLine(menu->disp, menu->win, menu->gc, menu->lbg, menu->hbg,
		   bd + GAP/2, ih + GAP/2, menu->width - 2 * bd - GAP);

    DrawString(menu->disp, menu->win, menu->font, menu->fg,
	       0, bd, menu->width, ih + GAP/2 - bd, menu->title, Center);
}

static void draw_item(MCPopupMenu* menu, int ix, int iy,
		      int iw, int ih, Bool selected_flag, const char* name)
{
    int bd = menu->bd_width;
    Pixel fg;
    int x, y;

    x = ix * iw + bd;
    y = iy * ih + bd + menu->title_offset;

    if(selected_flag)
	XSetForeground(menu->disp, menu->gc, menu->mark_color);
    else
	XSetForeground(menu->disp, menu->gc, menu->bg);

    XFillRectangle(menu->disp, menu->win, menu->gc,
		   x, y, iw, ih);
    if(selected_flag)
	DrawBorder(menu->disp, menu->win, menu->gc,
		   x, y, iw, ih, menu->bd_width, 
		   menu->lbg, menu->hbg, False);

    if(IsMonoDisplay(menu) && selected_flag)
	fg = menu->bg;
    else
	fg = menu->fg;

    DrawString(menu->disp, menu->win, menu->font, fg,
	       x + bd, y, iw - bd * 2, ih, name, West);
}

static void draw_all_menu_items(MCPopupMenu* menu,
				int item_width, 
				int item_height,
				int n, int nline)
{
    Window win = menu->win;
    Display* disp = menu->disp;
    int bd = menu->bd_width;
    int i;

    DrawBorder(disp, win, menu->gc,
	       0, 0, menu->width, menu->height, bd,
	       menu->lbg, menu->hbg, True);
    if(menu->title != NULL)
	draw_menu_title(menu, item_height);

    for(i = 0; i < n; i++)
    {
	int x, y;

	x = i / nline;
	y = i % nline;
	draw_item(menu, x, y, item_width, item_height, False,
		  menu->name_list[i]);
    }
}

Bool MCPopupMenu_popup_proc(WinObj* win_obj, int nline)
{
    MCPopupMenu* menu = (MCPopupMenu *)win_obj;
    int item_width, item_height;
    int n = menu->nitems;
    int w, h;
    int bd;
    Display* disp;
    Window win;
    int xmin, ymin, xmax, ymax;
    int title_width;

    if(win_obj->any.realized_flag == False)
	WinObjRealize(win_obj);

    if(n == 0 || menu->popup_flag == True)
	return False;
    item_width = popupmenubase_item_width_internal((PopupMenuBase *)menu)
	+ GAP;
    item_height = TextHeight(menu->font) + GAP + 2 * menu->bd_width;
    if(menu->title == NULL)
    {
	menu->title_offset = 0;
	title_width = 0;
    }
    else
    {
	menu->title_offset = item_height + GAP/2;
	title_width = TextWidth(menu->font, menu->title);
    }
    disp = menu->disp;
    win = menu->win;
    bd = menu->bd_width;

    if(nline < 0)
	nline = DisplayHeight(disp, DefaultScreen(disp)) / item_height;

    calc_win_size(n, item_width, item_height, nline, &w, &h);
    if(w + bd * 2 < title_width + GAP)
    {
	item_width = title_width / (w / item_width) + GAP;
	calc_win_size(n, item_width, item_height, nline, &w, &h);
    }

    ResizeWinObj((WinObj *)menu, w + bd * 2, h + bd * 2 + menu->title_offset);
    move_popup_window(menu, w + bd * 2, h + bd * 2 + menu->title_offset);
    XRaiseWindow(disp, win);
    XClearWindow(disp, win);
    MapWindow(disp, win, False);
    XFlush(disp);

    draw_all_menu_items(menu, item_width, item_height, n, nline);

    xmin = bd;
    ymin = menu->title_offset + bd;
    xmax = w + bd;
    ymax = h + menu->title_offset + bd;

    /* selecting loop */
    {
	XEvent e;
	int x, y, k0, k;
	Bool click_flag = True;
	int disp_width = DisplayWidth(disp, DefaultScreen(disp));
	int disp_height = DisplayHeight(disp, DefaultScreen(disp));

/*	XSetInputFocus(disp, win, RevertToParent, None); */
	if(XGrabPointer(disp, win, True,
			ButtonReleaseMask | ButtonMotionMask
			| PointerMotionMask,
			GrabModeAsync, GrabModeAsync, None, menu->cs, 0)
	   != GrabSuccess)
	{
	    CallWinObjErrorHandler(WOE_CANT_GRAB_POINTER, NULL);
	    XUnmapWindow(disp, win);
	    return False;
	}

	k0 = -1;
	menu->popup_flag = True;

#ifdef __linux__
	XSync(disp, False);
	usleep(100000);
#endif
	XSetInputFocus(disp, win, RevertToParent, None);
	while(1)
	{
	    if(menu->selecting_hook != NULL &&
	       menu->selecting_hook((PopupMenuBase *)menu) == 0)
	    {
		XSync(disp, False);
		if(QLength(disp) == 0)
		    continue;
	    }

	    do
	    {
		Bool redraw_flag;

		XNextEvent(disp, &e);

		switch(e.type)
		{
		  case ButtonRelease:
		    if(click_flag == True)
		    {
			click_flag = False;
			continue;
		    }
		    XUngrabPointer(disp, 0);
		    XUnmapWindow(disp, win);
		    XFlush(disp);
		    goto end_of_event_loop;

		  case MotionNotify:
		  case KeyPress:
		    if(e.type == KeyPress)
		    {
			if(popupmenubase_keypress_internal_proc(
				(PopupMenuBase *)menu, &e,
				item_width, item_height) == False)
			{
			    XUngrabPointer(disp, 0);
			    XUnmapWindow(disp, win);
			    XFlush(disp);
			    goto end_of_event_loop;
			}
			else
			{
			    x = e.xkey.x;
			    y = e.xkey.y;
			}
		    }
		    else
		    {
			click_flag = False;
			x = e.xmotion.x;
			y = e.xmotion.y;
		    }

		    redraw_flag = False;
		    if(menu->x < 0 && -x == menu->x)
		    {
			MoveWinObj((WinObj *)menu, menu->x + item_width,
				   menu->y);
			XWarpPointer(disp, None, win, 0, 0, 0, 0,
				     x, y);
			redraw_flag = True;
			draw_all_menu_items(menu, item_width, item_height,
					    n, nline);
		    }
		    else if(menu->x + menu->width >= disp_width &&
			    x + menu->x + 1 == disp_width)
		    {
			MoveWinObj((WinObj *)menu, menu->x - item_width,
				   menu->y);
			XWarpPointer(disp, None, win, 0, 0, 0, 0,
				     x, y);
			redraw_flag = True;
		    }
		    if(menu->y < 0 && -y == menu->y)
		    {
			MoveWinObj((WinObj *)menu, menu->x,
				   menu->y + item_height);
			XWarpPointer(disp, None, win, 0, 0, 0, 0,
				     x, y);
			redraw_flag = True;
			draw_all_menu_items(menu, item_width, item_height,
					    n, nline);
		    }
		    else if(menu->y + menu->height >= disp_height &&
			    y + menu->y + 1 == disp_height)
		    {
			MoveWinObj((WinObj *)menu, menu->x,
				   menu->y - item_height);
			XWarpPointer(disp, None, win, 0, 0, 0, 0,
				     x, y);
			redraw_flag = True;
		    }

		    if(redraw_flag == True)
			draw_all_menu_items(menu, item_width, item_height,
					    n, nline);

		    if(e.xmotion.window != menu->win ||
		       x < xmin || y < ymin || x >= xmax || y >= ymax)
		    {
			/* erase item */
			if(k0 != -1)
			{
			    draw_item(menu, k0 / nline, k0 % nline,
				      item_width, item_height, False,
				      menu->name_list[k0]);
			}
			k0 = -1;
			continue;
		    }

		    x = (x - xmin) / item_width;
		    y = (y - ymin) / item_height;
		    k = x * nline + y;
		    if(k >= n)
		    {
			/* erase item */
			if(k0 != -1)
			{
			    draw_item(menu, k0 / nline, k0 % nline,
				      item_width, item_height, False,
				      menu->name_list[k0]);
			}
			k0 = -1;
			continue;
		    }

		    if(k == k0) /* same item */
		    {
			if(redraw_flag == True)
			{
			    draw_item(menu, k0 / nline, k0 % nline,
				      item_width, item_height, True,
				      menu->name_list[k]);
			}

			continue;
		    }

		    if(k0 != -1)
		    {
			/* erase previous item */
			draw_item(menu, k0 / nline, k0 % nline,
				  item_width, item_height, False,
				  menu->name_list[k0]);
		    }
		    k0 = k;
		    /* draw item */
		    draw_item(menu, k / nline, k % nline,
			      item_width, item_height, True,
			      menu->name_list[k]);
		    break;
		}
	    } while(QLength(disp) > 0);
	}
      end_of_event_loop:
	menu->selected_item_index = k0;
	if(k0 >= 0)
	{
	    if(menu->selected_proc_list[k0] != NULL)
		menu->selected_proc_list[k0]((PopupMenuBase *)menu,
					     menu->name_list[k0],
					     menu->idlist[k0]);
	    menu->popup_flag = False;
	    return True;
	}
    }
    menu->popup_flag = False;
    return False;
}
