#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <string.h>
#include "WinObj.h"

#define MIN_MSG_DIALOG_WIDTH  100
#define OK_BUTTON_WIDTH  50
#define OK_BUTTON_HEIGHT 30
#define VGAP 30
#define HGAP 30

static void redraw_dialog(MessageDialog* md)
{
    DrawBorder(md->disp, md->win, md->gc,
	       0, 0, md->width, md->height,
	       md->bd_width, md->lbg, md->hbg, True);
    DrawBorder(md->disp, md->win, md->gc,
	       md->bd_width * 2, 
	       md->bd_width * 2,
	       md->width  - md->bd_width * 4,
	       md->height - md->bd_width * 4,
	       md->bd_width, md->lbg, md->hbg, False);

    DrawString(md->disp, md->win, md->font, md->fg,
	       0,
	       0,
	       md->width,
	       SkeletonWinObj(md->ok_button)->any.y, md->message, Center);
}

static void md_event_proc(WinObj* obj, XEvent* e);
static void restack_window(WinObj* obj)
{
    XEvent e;
    Display* disp = obj->any.disp;

    /* remove expose event */
    XSync(disp, False);
    while(XCheckTypedEvent(disp, Expose, &e) == True)
    {
	md_event_proc(obj, &e);
	XSync(disp, False);
    }

    /* raise window */
    XRaiseWindow(disp, obj->any.win);
    XSync(disp, False);
    if(XCheckTypedEvent(disp, Expose, &e) == True)
	md_event_proc(obj, &e);
    else
	XLowerWindow(disp, obj->any.win);
    XFlush(disp);
}

static void ok_button_proc(WinObj* obj, XEvent* e)
{
    MessageDialog* md = (MessageDialog *)obj->any.ext_data;
    XUnmapWindow(md->disp, md->win);
    md->mapped = False;
    winobj_free(md->message);
    md->message = NULL;
    XSync(md->disp, False);
}

static void md_event_proc(WinObj* obj, XEvent* e)
{
    static Window move_window_target = None;
    static int    move_window_x;
    static int    move_window_y;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count == 0 && obj->msg_dialog.message != NULL)
	    redraw_dialog((MessageDialog *)obj);
	break;
      case ButtonPress:
	switch(e->xbutton.button)
	{
	  case Button1:
	    if(e->xbutton.window == obj->any.win)
	    {
		XRaiseWindow(obj->any.disp, obj->any.win);
		XSync(obj->any.disp, False);
		move_window_target = e->xbutton.window;
		move_window_x = e->xbutton.x;
		move_window_y = e->xbutton.y;
	    }
	    break;
	  case Button3:
	    restack_window(obj);
	    break;
	}
	break;
      case ButtonRelease:
	move_window_target = None;
	break;
      case MotionNotify:
	if(e->xbutton.window == move_window_target)
	{
	    int x, y;

	    QueryPointer(obj->any.disp, DefaultRootWindow(obj->any.disp),
			 &x, &y);
	    MoveWinObj(obj, x - move_window_x, y - move_window_y);
	}
	break;
      case EnterNotify:
	XSetInputFocus(obj->any.disp, obj->any.win, RevertToParent, None);
	break;
      case LeaveNotify:
	XSetInputFocus(obj->any.disp, PointerRoot, RevertToPointerRoot, None);
	break;
      case KeyPress:
	if(XKeycodeToKeysym(obj->any.disp, e->xkey.keycode, 0) == XK_Return)
	    CallCallbackProc((WinObj *)obj->msg_dialog.ok_button, e);
	break;
    }
}

static void msgdialog_destruction(WinObj* obj)
{
    MessageDialog* md = (MessageDialog *)obj;
    winobj_free(md->message);
}

WinObj* create_MessageDialog(Display* disp,
			     int bd_width,
			     Pixel fg,
			     Pixel bg,
			     CharFont* font)
{
    MessageDialog* md = (MessageDialog *)
	toplevel_WinObj(disp, NULL, NULL, 1, 1, fg, bg);

    if(md == NULL)
	return NULL;

    md->ok_button = (ButtonWindow *)create_Button((WinObj *)md,
						  0,
						  0,
						  OK_BUTTON_WIDTH,
						  OK_BUTTON_HEIGHT,
						  bd_width,
						  font,
						  " OK ");
    if(md->ok_button == NULL)
    {
	DestroyWinObj((WinObj *)md);
	return NULL;
    }

    md->ok_button->ext_data = md;
    SetCallBackProc((WinObj *)md->ok_button, ok_button_proc);

    md->type = MessageDialog_T;
    md->font = font;
    md->bd_width = bd_width;
    md->when_realize_mapping = False;
    md->message = NULL;
    md->mapped = False;

    {
	XSetWindowAttributes at;
	at.override_redirect = True;
	at.save_under = True;
	XChangeWindowAttributes(md->disp, md->win,
				CWOverrideRedirect | CWSaveUnder, &at);
    }

    md->winobj_destruction = msgdialog_destruction;
    SetEventMask((WinObj *)md, ExposureMask | ButtonPressMask
		 | ButtonReleaseMask | ButtonMotionMask | KeyPressMask
		 | EnterWindowMask | LeaveWindowMask);

    SetEventHandler((WinObj *)md, md_event_proc);
    WinObjRealize((WinObj *)md);
    return (WinObj *)md;
}

void DisplayMessageDialog(WinObj* obj, const char* msg)
{
    int len;
    int str_width, str_height;
    MessageDialog* md = (MessageDialog *)obj;
    int win_width, win_height;
    int win_x, win_y;
    int px, py;
    int scr;

    if(msg == NULL)
	return;
    len = strlen(msg);
    if(len == 0)
	return;

    scr = DefaultScreen(md->disp);

    str_width  = TextWidth(md->font, msg);
    str_height = TextHeight(md->font);

    win_width  = MAX(MIN_MSG_DIALOG_WIDTH, str_width + 2 * HGAP);
    win_height = 3 * VGAP + str_height + OK_BUTTON_HEIGHT;

    MoveWinObj((WinObj*)md->ok_button,
	       (win_width - OK_BUTTON_WIDTH) / 2,
	       2 * VGAP + str_height);

    QueryPointer(md->disp, DefaultRootWindow(md->disp), &px, &py);
    win_x = px - win_width / 2;
    win_y = py - (2 * VGAP + str_height + OK_BUTTON_HEIGHT/2);

    if(win_x < 0)
	win_x = 0;
    else if(win_x + win_width > DisplayWidth(md->disp, scr))
	win_x = DisplayWidth(md->disp, scr) - win_width;

    if(win_y < 0)
	win_y = 0;
    else if(win_y + win_height > DisplayHeight(md->disp, scr))
	win_y = DisplayHeight(md->disp, scr) - win_height;

    ResizeWinObj(obj, win_width, win_height);
    XRaiseWindow(md->disp, md->win);
    if(md->mapped == False)
    {
	MoveWinObj(obj, win_x, win_y);
	XWarpPointer(md->disp, None, md->ok_button->win,
		     None, None, None, None,
		     OK_BUTTON_WIDTH / 2, OK_BUTTON_HEIGHT / 2);
	XMapWindow(md->disp, md->win);
    }
    md->mapped = True;

    if((md->message = winobj_strdup(msg)) == NULL)
    {
	fprintf(stderr, "Can't allocate memory.\n");
	md->message = winobj_null_string;
    }

    XSync(md->disp, False);
}
