#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <string.h>
#include "WinObj.h"

#define INIT_MENU_SIZE 64
#define GAP 8
#define ITEM_TRIG_MARK_SIZE 8
#define ITEM_SIDE_MARGIN 12
#define MARKER_MARGIN 8

static void popup_map_proc(WinObj* win_obj, int x, int y);
static void popup_unmap_proc(WinObj* win_obj);
static void draw_item_proc(WinObj* obj, int iy, Bool selected_flag);

static void popup_destruction(WinObj* obj)
{
    PopupMenu* menu = (PopupMenu *)obj;
    int i;

    winobj_free(menu->title);
    for(i = 0; i < menu->nitems; i++)
	winobj_free(menu->name_list[i]);
    winobj_free(menu->name_list);
    winobj_free(menu->idlist);
    winobj_free(menu->selected_proc_list);
    winobj_free(menu->hierarchical);
    winobj_free(menu->marker);
    winobj_free(menu->noselect);
}

WinObj* create_PopupMenu(Display* disp,
			 const char* title,
			 int bd_width,
			 Pixel fg,
			 Pixel bg,
			 CharFont* font)
{
    PopupMenu* menu = (PopupMenu *)
	create_PopupMenuBase(disp, title, bd_width, fg, bg, font);

    if(menu == NULL)
	return NULL;

    menu->type = PopupMenu_T;
    menu->hierarchical = NULL;
    menu->marker = NULL;
    menu->noselect = NULL;
    menu->hierarchical_target = NULL;
    menu->popup_background = None;
    menu->item_height = TextHeight(menu->font) + GAP + 2 * bd_width;
    menu->popup_map_proc   = popup_map_proc;
    menu->popup_unmap_proc = popup_unmap_proc;
    menu->draw_item_proc = draw_item_proc;
    menu->update_flag = False;

    if(title == NULL)
    {
	menu->title_offset = 0;
	menu->width = GAP + 2 * bd_width + 2 * ITEM_SIDE_MARGIN;
	menu->height = 2 * bd_width;
    }
    else
    {
	menu->title_offset = menu->item_height + GAP/2;
	menu->width = GAP + 2 * bd_width + TextWidth(font, title)
	    + 2 * ITEM_SIDE_MARGIN;
	menu->height = 2 * bd_width + menu->title_offset;
    }

    menu->bar_ref = NULL;
    menu->winobj_destruction = popup_destruction;

    SetEventMask((WinObj *)menu, PointerMotionMask |
		 ButtonPressMask | ButtonReleaseMask |
		 LeaveWindowMask | EnterWindowMask);

    return (WinObj*)menu;
}

Bool add_PopupMenu(WinObj* win_obj,
		   const char* name,
		   int menu_id,
		   void (*selected_proc)
		   		(PopupMenuBase* menu, char* name, int id))
{
    PopupMenu* menu = (PopupMenu *)win_obj;
    int i, k = menu->nitems;
    int name_width;

    if(menu->name_list == NULL)
    {
	menu->name_list =
	    (char **)winobj_malloc(INIT_MENU_SIZE * sizeof(char *));
	menu->idlist =
	    (int *)winobj_malloc(INIT_MENU_SIZE * sizeof(int));
	menu->selected_proc_list = (void (**)(PopupMenuBase *, char *, int))
	    winobj_malloc(INIT_MENU_SIZE * sizeof(void (*)(char *, int)));
	menu->hierarchical = (PopupMenu **)
	    winobj_malloc(INIT_MENU_SIZE * sizeof(PopupMenu *));
	menu->marker = (char *)winobj_malloc(INIT_MENU_SIZE);
	menu->noselect = (Bool *)winobj_malloc(INIT_MENU_SIZE * sizeof(Bool));
	if(menu->name_list == NULL ||
	   menu->idlist == NULL ||
	   menu->selected_proc_list == NULL ||
	   menu->hierarchical == NULL ||
	   menu->marker == NULL ||
	   menu->noselect == NULL)
	{
	    winobj_free(menu->name_list);
	    winobj_free(menu->idlist);
	    winobj_free(menu->selected_proc_list);
	    winobj_free(menu->hierarchical);
	    winobj_free(menu->marker);
	    winobj_free(menu->noselect);
	    menu->name_list = NULL;
	    return False;
	}

	for(i = 0; i < INIT_MENU_SIZE; i++)
	{
	    menu->hierarchical[i] = NULL;
	    menu->marker[i] = False;
	    menu->noselect[i] = False;
	}
    }

    menu->update_flag = True;
    if(k == menu->maxitem)
    {
#define REALLOC(obj, type, newsize) realloc_ptr = type winobj_realloc(obj, newsize); if(realloc_ptr == NULL) return False; else obj = type realloc_ptr;

	int n = menu->maxitem * 2, i;
	void* realloc_ptr;

	REALLOC(menu->name_list, (char **), n * sizeof(char *));
	REALLOC(menu->idlist, (int *), n * sizeof(int));
	REALLOC(menu->selected_proc_list, (void (**)(PopupMenuBase*,char*,int))
		, n * sizeof(void (*)(char*,int)));
	REALLOC(menu->hierarchical, (PopupMenu **), n * sizeof(PopupMenu *));
	REALLOC(menu->marker, (char *), n);
	REALLOC(menu->noselect, (Bool *), n * sizeof(Bool));

	menu->maxitem = n;
	for(i = k; i < n; i++)
	{
	    menu->hierarchical[i] = NULL;
	    menu->marker[i] = False;
	    menu->noselect[i] = False;
	}
#undef REALLOC
    }

    if(name == NULL)
	menu->name_list[k] = NULL;
    else
    {
	if((menu->name_list[k] = winobj_strdup(name)) == NULL)
	    return False;
	name_width = TextWidth(menu->font, name) + 2 * GAP
	    + 4 * menu->bd_width + 2 * ITEM_SIDE_MARGIN;
	if(name_width  > menu->width)
	    menu->width = name_width;
    }
    menu->idlist[k] = menu_id;
    menu->selected_proc_list[k] = selected_proc;
    menu->nitems = k + 1;
    menu->height += menu->item_height;
    return True;
}

void clear_PopupMenu(WinObj* win_obj)
{
    PopupMenu* menu = (PopupMenu *)win_obj;
    int bd_width = menu->bd_width;

    menu->update_flag = True;
    clear_PopupMenuBase(win_obj);
    if(menu->title == NULL)
    {
	menu->title_offset = 0;
	menu->width = GAP + 2 * bd_width + 2 * ITEM_SIDE_MARGIN;
	menu->height = 2 * bd_width;
    }
    else
    {
	menu->title_offset = menu->item_height + GAP/2;
	menu->width = GAP + 2 * bd_width + TextWidth(menu->font, menu->title)
	    + 2 * ITEM_SIDE_MARGIN;
	menu->height = 2 * bd_width + menu->title_offset;
    }
}

static void fit_width(PopupMenu* menu)
{
    int i, width;

    if(menu->title == NULL)
	width = 0;
    else
	width = TextWidth(menu->font, menu->title);

    for(i = 0; i < menu->nitems; i++)
    {
	int w = TextWidth(menu->font, menu->name_list[i]);
	if(w > width)
	    width = w;
    }

    menu->width = GAP + 2 * menu->bd_width + width + 2 * ITEM_SIDE_MARGIN;
}

Bool rename_PopupMenu(WinObj* win_obj, const char* name, int id)
{
    win_obj->popup_menu.update_flag = True;
    if(rename_PopupMenuBase(win_obj, name, id))
    {
	fit_width((PopupMenu *)win_obj);
	return True;
    }
    return False;
}

static void move_popup_window(PopupMenu* menu, int x, int y)
{
    Display* disp = menu->disp;
    int scr = DefaultScreen(disp);
    int w = menu->width;
    int h = menu->height;

    if(x < 0)
	x = 0;
    if(y < 0)
	y = 0;

    else if(x + w >= DisplayWidth(disp, scr))
	x = DisplayWidth(disp, scr) - w - 1;

    if(y + h >= DisplayHeight(disp, scr))
	y = DisplayHeight(disp, scr) - h - 1;
    MoveWinObj((WinObj *)menu, x, y);
}

static void draw_menu_title(PopupMenu* menu, int ih)
{
    int bd = menu->bd_width;
    DrawRuledHLine(menu->disp, menu->win, menu->gc, menu->lbg, menu->hbg,
		   bd + GAP/2, ih + GAP/2, menu->width - 2 * bd - GAP);

    DrawString(menu->disp, menu->win, menu->font, menu->fg,
	       0, bd, menu->width, ih + GAP/2 - bd, menu->title, Center);
}

static void draw_item(PopupMenu* menu, int n, Bool selected_flag)
{
    menu->draw_item_proc((WinObj *)menu, n, selected_flag);
}

static void draw_item_proc(WinObj* obj, int iy, Bool selected_flag)
{
    PopupMenu* menu = (PopupMenu *)obj;
    int bd = menu->bd_width;
    Pixel fg;
    int x, y;
    int iw = menu->width - 2 * bd;
    int ih = menu->item_height;
    char* name = menu->name_list[iy];

    x = bd;
    y = iy * ih + bd + menu->title_offset;

    if(selected_flag && name != NULL)
	XSetForeground(menu->disp, menu->gc, menu->mark_color);
    else
	XSetForeground(menu->disp, menu->gc, menu->bg);

    XFillRectangle(menu->disp, menu->win, menu->gc, x, y, iw, ih);

    if(selected_flag && name != NULL)
	DrawBorder(menu->disp, menu->win, menu->gc,
		   x, y, iw, ih, menu->bd_width, 
		   menu->lbg, menu->hbg, True);

    if(IsMonoDisplay(menu) && selected_flag)
	fg = menu->bg;
    else
	fg = menu->fg;

    if(name != NULL)
    {
	DrawString(menu->disp, menu->win, menu->font, fg,
		   ITEM_SIDE_MARGIN + 2 * bd + GAP/2, y, 0, ih, name, West);
	if(menu->marker[iy])
	{
	    DrawBorder(menu->disp, menu->win, menu->gc,
		       ITEM_SIDE_MARGIN + 2 * bd + GAP/2 - MARKER_MARGIN,
		       y + (ih - TextHeight(menu->font)) / 2,
		       TextWidth(menu->font, name) + 2 * MARKER_MARGIN,
		       TextHeight(menu->font),
		       1, menu->lbg, menu->hbg, selected_flag);
	    DrawBorder(menu->disp, menu->win, menu->gc,
		       ITEM_SIDE_MARGIN + 2 * bd + GAP/2 - 3 - MARKER_MARGIN,
		       y + (ih - TextHeight(menu->font)) / 2 - 3,
		       TextWidth(menu->font, name) + 6 + 2 * MARKER_MARGIN,
		       TextHeight(menu->font) + 6,
		       1, menu->lbg, menu->hbg, !selected_flag);
	}
    }
    else
    {
	DrawRuledHLine(menu->disp, menu->win, menu->gc,
		       menu->lbg, menu->hbg,
		       bd + GAP / 2,
		       y + ih / 2,
		       menu->width - 2 * bd - GAP - 1);
    }

    if(menu->hierarchical[iy] != NULL)
    {
	Pixel bg;

	if(IsMonoDisplay(menu))
	    if(selected_flag)
		bg = menu->fg;
	    else
		bg = menu->bg;
	else
	    if(selected_flag)
		bg = menu->mark_color;
	    else
		bg = menu->bg;

	DrawTriangle(menu->disp, menu->win, menu->gc,
		     x + iw - bd - ITEM_TRIG_MARK_SIZE - 1,
		     y + ih / 2,
		     1, (double)ITEM_TRIG_MARK_SIZE, 0.0,
		     bg, menu->lbg, menu->hbg, !selected_flag);
    }
    if(menu->noselect[iy])
    {
	XSetForeground(menu->disp, menu->gc, menu->bg);
	DrawChecker(menu->disp, menu->win, menu->gc, x, y, iw, ih);
    }
}

static void draw_all_menu_items(PopupMenu* menu, Bool initialize_background)
{
    Window win = menu->win;
    Display* disp = menu->disp;

    if(initialize_background)
    {
	int bd = menu->bd_width;
	int i;
	int item_height = menu->item_height;
	int n = menu->nitems;

	menu->win = menu->popup_background;
	DrawBorder(disp, menu->popup_background, menu->gc,
		   0, 0, menu->width, menu->height, bd,
		   menu->lbg, menu->hbg, True);
	if(menu->title != NULL)
	    draw_menu_title(menu, item_height);
	for(i = 0; i < n; i++)
	    draw_item(menu, i, False);
	menu->win = win;
	draw_all_menu_items(menu, False);
    }
    else
	XCopyArea(disp, menu->popup_background, win, menu->gc,
		  0, 0, menu->width, menu->height, 0, 0);
}

static void popup_map_proc(WinObj* win_obj, int x, int y)
{
    PopupMenu* menu = (PopupMenu *)win_obj;
    Display* disp = menu->disp;
    Window win = menu->win;

    if(win_obj->any.realized_flag == False)
	WinObjRealize(win_obj);

    ResizeWinObj((WinObj *)menu, menu->width, menu->height);
    move_popup_window(menu, x, y);
    XRaiseWindow(disp, win);
    XClearWindow(disp, win);
    MapWindow(disp, win, False);

    if(menu->update_flag)
    {
	if(menu->popup_background != None)
	    XFreePixmap(disp, menu->popup_background);
	menu->popup_background =
	    CreatePixmap(disp, win, menu->width, menu->height, menu->bg);
	draw_all_menu_items(menu, True);
    }
    else
    {
	if(menu->popup_background == None)
	{
	    menu->popup_background =
		CreatePixmap(disp, win, menu->width, menu->height, menu->bg);
	    draw_all_menu_items(menu, True);
	}
	else
	    draw_all_menu_items(menu, False);
    }
    XFlush(disp);
}

static void popup_menu_list_default(PopupMenu* menu)
{
    int x, y;

    QueryRootPointer(menu->disp, &x, &y);
    y -= menu->item_height / 2;
    x -= menu->width / 2;
    if(menu->realized_flag == False)
	WinObjRealize((WinObj *)menu);
    menu->popup_map_proc((WinObj *)menu, x, y);
}


static void popup_unmap_proc(WinObj* obj)
{
    PopupMenu* menu = (PopupMenu *)obj;
    if(menu->hierarchical_target)
	unmap_popup_menu_internal(menu->hierarchical_target);
    XUnmapWindow(menu->disp, menu->win);
    menu->popup_flag = False;
    menu->hierarchical_target = NULL;
    menu->selected_item_index = -1;
}

void unmap_popup_menu_internal(PopupMenu* menu)
{
    menu->popup_unmap_proc((WinObj *)menu);
}

PopupMenu* dispatch_current_popupmenu(PopupMenu* menu_root, Window win)
{
    PopupMenu* p;

    for(p = menu_root; p; p = p->hierarchical_target)
	if(p->win == win)
	    break;

    return p;
}

static void switch_menu(PopupMenu* menu, int k)
{
    if(k == menu->selected_item_index)
	return;

    if(menu->selected_item_index != -1 &&
       menu->hierarchical_target == NULL)
	draw_item(menu, menu->selected_item_index, False);

    if(k != -1)
    {
	draw_item(menu, k, True);
	menu->selected_item_index = k;
    }
    else if (menu->hierarchical_target == NULL)
	menu->selected_item_index = k;
}

static int calc_popup_menu_index(PopupMenu* menu, int x, int y)
{
    int bd = menu->bd_width;
    int idx;

    if(x < bd ||
       x >= menu->width - bd ||
       y < menu->title_offset + bd ||
       y >= menu->nitems * menu->item_height + menu->title_offset + bd)
	idx = -1;
    else
    {
	idx = (y - menu->title_offset - bd) / menu->item_height;
	if(menu->noselect[idx])
	    idx = -1;
    }

    return idx;
}


Bool PopupMenu_popup_proc(WinObj* win_obj, int x, int y)
{
    PopupMenu* menu = (PopupMenu *)win_obj;
    PopupMenu* menu_root = menu;
    MenuBar* bar = menu->bar_ref;

    if(!WinObjPopupAble(win_obj))
    {
	fprintf(stderr,
"WinObj Fatal Error: Invalid WinObj is specified in PopupMenu_popup_proc()\n");
	exit(1);
    }

    if(menu->nitems == 0 || menu->popup_flag == True)
	return False;

    if(menu->realized_flag == False)
	WinObjRealize(win_obj);

    menu->popup_map_proc((WinObj *)menu, x, y);

    /* selecting loop */
    {
	XEvent e;
	int x, y, k = -1;
	Bool click_flag = True;
	Display* disp = menu->disp;
	int disp_height = DisplayHeight(disp, DefaultScreen(disp));

	XRaiseWindow(disp, menu->win);
	if(XGrabPointer(disp, menu->win, True,
		     ButtonReleaseMask | ButtonMotionMask | PointerMotionMask,
		     GrabModeAsync, GrabModeAsync, None, menu->cs, 0)
	   != GrabSuccess)
	{
	    CallWinObjErrorHandler(WOE_CANT_GRAB_POINTER, NULL);
	    XUnmapWindow(disp, menu->win);
	    return False;
	}

	menu->popup_flag = True;
	menu->selected_item_index = -1;

	while(1)
	{
	    Bool (* selecting_hook)(struct _PopupMenuBase *) =
		menu->selecting_hook;

	    if(selecting_hook != NULL &&
	       selecting_hook((PopupMenuBase *)menu) == 0)
	    {
		XSync(disp, False);
		if(QLength(disp) == 0)
		    continue;
	    }

	    do
	    {
		Bool redraw_flag;
		int bd;

		XNextEvent(disp, &e);

		if(bar != NULL &&
		   bar->win == e.xany.window)
		{
		    int status;
		    if(bar->switch_menu_proc(bar, &e, &status) == False)
			return status;
		    continue;
		}

		if(e.type == Expose)
		{
		    DispatchEvent(&e);
		    continue;
		}

		if(e.type == MotionNotify ||
		   e.type == LeaveNotify ||
		   e.type == EnterNotify)
		{
		    PopupMenu* m;

		    m = dispatch_current_popupmenu(menu_root, e.xany.window);

		    if(m == NULL)
			continue;
		    menu = m;
		}

		bd = menu->bd_width;
		redraw_flag = False;
		switch(e.type)
		{
		  case LeaveNotify:
		    switch_menu(menu, -1);
		    break;

		  case EnterNotify:
		    x = e.xcrossing.x;
		    y = e.xcrossing.y;

		    k = calc_popup_menu_index(menu, x, y);
		    if(k == menu->selected_item_index)
			continue;

		    if(menu->hierarchical_target != NULL)
		    {
			unmap_popup_menu_internal(menu->hierarchical_target);
			draw_all_menu_items(menu, False);
			menu->hierarchical_target = NULL;
			draw_item(menu, menu->selected_item_index, True);
		    }

		    goto pointer_motion_event;

		  case ButtonRelease:
		    if(click_flag == True)
		    {
			click_flag = False;
			continue;
		    }
		    goto end_of_event_loop;

		  case MotionNotify:
		    click_flag = False;
		    x = e.xmotion.x;
		    y = e.xmotion.y;
		    if(menu->y < 0 && -y == menu->y)
		    {
			MoveWinObj((WinObj *)menu, menu->x,
				   menu->y + menu->item_height);
			XWarpPointer(disp, None, menu->win, 0, 0, 0, 0,
				     x, y);
			redraw_flag = True;
			draw_all_menu_items(menu, False);
		    }
		    else if(menu->y + menu->height >= disp_height &&
			    y + menu->y + 1 == disp_height)
		    {
			MoveWinObj((WinObj *)menu, menu->x,
				   menu->y - menu->item_height);
			XWarpPointer(disp, None, menu->win, 0, 0, 0, 0,
				     x, y);
			redraw_flag = True;
		    }

		    if(redraw_flag == True)
			draw_all_menu_items(menu, False);

		    k = calc_popup_menu_index(menu, x, y);

		  pointer_motion_event:
		    if(k == -1)
		    {
			/* erase item */
			switch_menu(menu, -1);
			continue;
		    }

		    if(menu->selected_item_index != k &&
		       menu->hierarchical[menu->selected_item_index] != NULL &&
		       menu->hierarchical[menu->selected_item_index] ==
			       menu->hierarchical_target)
		       
		    {
			unmap_popup_menu_internal(menu->hierarchical_target);
			draw_all_menu_items(menu, False);
			menu->hierarchical_target = NULL;
		    }

		    switch_menu(menu, k);
		    if(menu->hierarchical[k] != NULL &&
		       menu->hierarchical[k]->popup_flag == False &&
		       x < menu->width &&
		       x > menu->width - 2 * ITEM_TRIG_MARK_SIZE - 2 * bd)
		    {
			PopupMenu* m;

			m = menu->hierarchical[k];
			popup_menu_list_default(m);
			m->popup_flag = True;

			if(menu->hierarchical_target != NULL)
			{
			    unmap_popup_menu_internal(
				menu->hierarchical_target);
			    draw_all_menu_items(menu, False);
			    draw_item(menu, k, True);
			}
			menu->hierarchical_target = m;
			XSync(disp, False);
			menu = m;

			continue;
		    }
		    break;
		}
	    } while(QLength(disp) > 0);
	} /* while (1) */
      end_of_event_loop:
	{
/*	    int k = menu->selected_item_index; */
	    XUngrabPointer(menu_root->disp, 0);
	    unmap_popup_menu_internal(menu_root);
	    menu->selected_item_index = k;
	    XFlush(menu_root->disp);

	    if(k >= 0)
	    {
		if(menu->selected_proc_list[k] != NULL)
		{
		    menu->selected_proc_list[k]((PopupMenuBase *)menu,
						menu->name_list[k],
						menu->idlist[k]);
		}
		return True;
	    }
	}
    }
    return False;
}

void PopupMenuHierarchicalHook(PopupMenuBase* popup, char* name, int id)
{
    PopupMenu* menu = (PopupMenu *)popup;
    int i;

    i = PopupMenuIdToIndex((WinObj *)popup, id, True);
    if(menu->hierarchical[i] != NULL)
    {
	int x, y;
	QueryRootPointer(menu->disp, &x, &y);
	y -= menu->item_height / 2;
	x -= menu->width / 2;
	PopupMenu_popup_proc((WinObj *)menu->hierarchical[i], x, y);
    }
}

void PopupMenuSetNextMenu(WinObj* popup, int id, WinObj* next_menu)
{
    int n;

    PopupMenu* menu = (PopupMenu *)popup;

    menu->update_flag = True;
    if(!WinObjPopupAble(popup) || !WinObjPopupAble(next_menu))
    {
	fprintf(stderr,
"WinObj Fatal Error: Invalid WinObj data specified in PopupMenuSetNextMenu()\n"
		);
	exit(1);
    }

    for(n = 0; n < menu->nitems; n++)
	if(menu->hierarchical[n])
	    break;

    n = PopupMenuIdToIndex(popup, id, True);
    menu->hierarchical[n] = (PopupMenu *)next_menu;
}

void PopupMenuNoselect(WinObj* obj, int id)
{
    PopupMenu* menu = (PopupMenu *)obj;
    int idx;

    if(!WinObjPopupAble(obj))
    {
	fprintf(stderr,
"WinObj Fatal Error: Invalid WinObj data specified in PopupMenuNoselect()\n");
	exit(1);
    }

    idx = PopupMenuIdToIndex(obj, id, True);
    menu->noselect[idx] = True;
}

void PopupMenuResetNoselect(WinObj* obj, int id)
{
    PopupMenu* menu = (PopupMenu *)obj;
    int idx;

    if(!WinObjPopupAble(obj))
    {
	fprintf(stderr,
"WinObj Fatal Error: Invalid WinObj data specified in PopupMenuResetNoselect()\n");
	exit(1);
    }
    
    idx = PopupMenuIdToIndex(obj, id, True);
    menu->noselect[idx] = False;
}

int PopupMenuChangeMarker(WinObj* obj, int new_id)
{
    PopupMenu* menu = (PopupMenu *)obj;
    int i, idx;

    if(!WinObjPopupAble(obj))
    {
	fprintf(stderr,
"WinObj Fatal Error: Invalid WinObj data specified in PopupMenuChangeMarker()\n");
	exit(1);
    }

    idx = PopupMenuIdToIndex(obj, new_id, True);
    for(i = 0; i < menu->nitems; i++)
    {
	if(menu->marker[i] == True)
	{
	    menu->marker[i] = False;
	    menu->marker[idx] = True;
	    return menu->idlist[i];
	}
    }
    menu->marker[idx] = True;
    return menu->idlist[0];	/* Not found */
}

int PopupMenuMarkerID(WinObj* obj)
{
    PopupMenu* menu = (PopupMenu *)obj;
    int i;

    if(!WinObjPopupAble(obj))
    {
	fprintf(stderr,
"WinObj Fatal Error: Invalid WinObj data specified in PopupMenuChangeMarker()\n");
	exit(1);
    }

    for(i = 0; i < menu->nitems; i++)
	if(menu->marker[i])
	    return menu->idlist[i];
    return menu->idlist[0];	/* Not found */
}
