#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <string.h>
#include "WinObj.h"

#define INIT_MENU_SIZE 64

int PopupMenuIdToIndex(WinObj* popup, int id, Bool errorp)
{
    int i;
    PopupMenuBase* menu = (PopupMenuBase *)popup;

    for(i = 0; i < menu->nitems; i++)
	if(menu->idlist[i] == id)
	    return i;

    if(errorp)
    {
	fprintf(stderr,
		"WinObj Fatal Error: Undefined PopupMenu id: %d\n", id);
	exit(1);
    }
    return -1;
}

static void popupmenubase_destruction(WinObj* obj)
{
    PopupMenuBase* menu = (PopupMenuBase *)obj;
    int i;

    if(menu->title != NULL)
	winobj_free(menu->title);
    for(i = 0; i < menu->nitems; i++)
	if(menu->name_list[i] != NULL)
	    winobj_free(menu->name_list[i]);
    winobj_free(menu->name_list);
    winobj_free(menu->idlist);
    winobj_free(menu->selected_proc_list);
}

WinObj* create_PopupMenuBase(Display* disp,
			     const char* title,
			     int bd_width,
			     Pixel fg,
			     Pixel bg,
			     CharFont* font)
{
    Window win;
    PopupMenuBase* menu = (PopupMenuBase *)
	toplevel_WinObj(disp, NULL, NULL, 1, 1, fg, bg);

    if(menu == NULL)
	return NULL;

    win = menu->win;
    menu->type = PopupMenuBase_T;
    menu->when_realize_mapping = False;
    menu->bd_width = bd_width;
    menu->font = font;
    menu->name_list = NULL;
    menu->idlist = NULL;
    menu->selected_proc_list = NULL;
    menu->nitems = 0;
    menu->maxitem = INIT_MENU_SIZE;
    menu->selected_item_index = -1;
    menu->selecting_hook = NULL;

    SetEventMask((WinObj *)menu, PointerMotionMask | ButtonPressMask |
		 ButtonReleaseMask | KeyPressMask);

    {
	XSetWindowAttributes at;
	at.override_redirect = True;
	at.save_under = True;
	XChangeWindowAttributes(disp, win,
				CWOverrideRedirect | CWSaveUnder, &at);
    }

    menu->cs = XCreateFontCursor(disp, XC_hand2);
    XDefineCursor(disp, win, menu->cs);

    if(title != NULL)
	menu->title = winobj_strdup(title);
    else
	menu->title = NULL;

    menu->popup_flag = False;
    menu->winobj_destruction = popupmenubase_destruction;

    return (WinObj*)menu;
}

Bool add_PopupMenuBase(WinObj* win_obj,
		       const char* name,
		       int menu_id,
		       void (*selected_proc)
		       		(PopupMenuBase* m, char* name, int id))
{
    PopupMenuBase* menu = (PopupMenuBase *)win_obj;
    int k = menu->nitems;

    if(menu->name_list == NULL)
    {
	menu->name_list =
	    (char **)winobj_malloc(INIT_MENU_SIZE * sizeof(char *));
	menu->idlist =
	    (int *)winobj_malloc(INIT_MENU_SIZE * sizeof(int));
	menu->selected_proc_list = (void (**)(PopupMenuBase *, char *, int))
	    winobj_malloc(INIT_MENU_SIZE * sizeof(void (*)(char *, int)));

	if(menu->name_list == NULL ||
	   menu->idlist == NULL ||
	   menu->selected_proc_list == NULL)
	{
	    winobj_free(menu->name_list);
	    winobj_free(menu->idlist);
	    winobj_free(menu->selected_proc_list);
	    menu->name_list = NULL;
	    return False;
	}
    }

    if(k == menu->maxitem)
    {
#define REALLOC(obj, type, newsize) realloc_ptr = type winobj_realloc(obj, newsize); if(realloc_ptr == NULL) return False; else obj = type realloc_ptr;

	int n = menu->maxitem * 2;
	void* realloc_ptr;

	REALLOC(menu->name_list, (char **), n * sizeof(char *));
	REALLOC(menu->idlist, (int *), n * sizeof(int));
	REALLOC(menu->selected_proc_list, (void (**)(PopupMenuBase*,char*,int))
		, n * sizeof(void (*)(char*,int)));
	menu->maxitem = n;
#undef REALLOC
    }

    if((menu->name_list[k] = winobj_strdup(name)) == NULL)
	return False;
    menu->idlist[k] = menu_id;
    menu->selected_proc_list[k] = selected_proc;
    menu->nitems = k + 1;
    return True;
}

void clear_PopupMenuBase(WinObj* menu)
{
    int i, n;

    n = ((PopupMenuBase *)menu)->nitems;
    for(i = 0; i < n; i++)
    {
	if(((PopupMenuBase *)menu)->name_list[i] != NULL)
	    winobj_free(((PopupMenuBase *)menu)->name_list[i]);
    }
    ((PopupMenuBase *)menu)->nitems = 0;
}

Bool popupmenubase_keypress_internal_proc(PopupMenuBase* menu,
					  XEvent* e, int iw, int ih)
{
    KeySym sym;
    int x, y;

    sym = XKeycodeToKeysym(e->xkey.display, e->xkey.keycode, 0);
    QueryPointer(menu->disp, menu->win, &x, &y);
    x -= menu->x;
    y -= menu->y;

    switch(sym)
    {
      case XK_Return:
#ifdef XK_KP_Space
      case XK_KP_Space:
#endif
#ifdef XK_space
      case XK_space:
#endif
	return False;

      case XK_n:
#ifdef XK_Down
      case XK_Down:
#endif
#ifdef XK_KP_Down
      case XK_KP_Down:
#endif
	y += ih;
	break;

      case XK_p:
#ifdef XK_Up
      case XK_Up:
#endif
#ifdef XK_KP_Up
      case XK_KP_Up:
#endif
	y -= ih;
	break;

      case XK_b:
#ifdef XK_Left
      case XK_Left:
#endif
#ifdef XK_KP_Left
      case XK_KP_Left:
#endif
	x -= iw;
	break;

      case XK_f:
#ifdef XK_Right
      case XK_Right:
#endif
#ifdef XK_KP_Left
      case XK_KP_Right:
#endif
	x += iw;
	break;
    }

    XWarpPointer(menu->disp, None, menu->win, 0, 0, 0, 0, x, y);
    QueryPointer(menu->disp, menu->win, &x, &y);
    x -= menu->x;
    y -= menu->y;
    e->xkey.x = x;
    e->xkey.y = y;

    return True;
}

Bool rename_PopupMenuBase(WinObj* win_obj, const char* name, int id)
{
    PopupMenuBase* menu = (PopupMenuBase *)win_obj;
    int n;

    n = PopupMenuIdToIndex(win_obj, id, True);
    if(menu->name_list[n] != NULL)
	winobj_free(menu->name_list[n]);
    if(name == NULL)
	menu->name_list[n] = NULL;
    else
    {
	if((menu->name_list[n] = winobj_strdup(name)) == NULL)
	{
	    menu->name_list[n] = winobj_null_string;
	    return False;
	}
    }
    return True;
}

int popupmenubase_item_width_internal(PopupMenuBase* menu)
{
    int i, n, maxlen, len;

    n = menu->nitems;
    maxlen = 0;
    for(i = 0; i < n; i++)
    {
	len = TextWidth(menu->font, menu->name_list[i]);
	if(len > maxlen)
	    maxlen = len;
    }

    return maxlen + 2 * menu->bd_width;
}
