#include <math.h>
#include "colcnv.h"
#include "winobj_defs.h"

static double h2v(double h, double cmin, double cmax)
{
    h = fmod(h, 360.0);
    if (h < 0)
	h += 360;

    if (h < 60)
	return cmin + (cmax - cmin) * h / 60.0;
    if (h >= 60 && h < 180)
	return cmax;
    if (h >= 180 && h < 240)
	return cmin + (cmax - cmin) * (240 - h) / 60.0;

    return cmin;
}

/******************** RGB <---> HSV ********************/
void rgb_to_hsv(rgb_t* rgb, double* h, double* s, double* v)
{
    double cmax, cmin, cd;

    cmax = MAX3(rgb->r, rgb->g, rgb->b);
    cmin = MIN3(rgb->r, rgb->g, rgb->b);
/*
    if (rgb->r >= rgb->g)
	cmax = rgb->r;
    else
	cmax = rgb->g;
    if (rgb->b > cmax)
	cmax = rgb->b;

    if (rgb->r <= rgb->g)
	cmin = rgb->r;
    else
	cmin = rgb->g;
    if (rgb->b < cmin)
	cmin = rgb->b; */

    *v = cmax;
    cd = cmax - cmin;

    if (cmax == 0)
	*s = 0;
    else
	*s = cd / cmax;

    if (*s == 0) {
	*h = 0;
	return;
    }

    if (rgb->r == cmax)
	*h = (rgb->g - rgb->b) / cd;
    else if (rgb->g == cmax)
	*h = 2 + (rgb->b - rgb->r) / cd;
    else if (rgb->b == cmax)
	*h = 4 + (rgb->r - rgb->g) / cd;

    *h *= 60;
    if (*h < 0)
	*h += 360;
}

void hsv_to_rgb(double h, double s, double v, rgb_t* rgb)
{
    double f;
    double i;
    double p1, p2, p3;

    if (s < 0)
	s = 0;
    if (v < 0)
	v = 0;

    if (s > 1)
	s = 1;
    if (v > 1)
	v = 1;

    h = fmod(h, 360.0);
    if (h < 0)
	h += 360;

    h /= 60;
    f = modf(h, &i);
    p1 = v * (1 - s);
    p2 = v * (1 - s * f);
    p3 = v * (1 - s * (1 - f));

    switch ((int)i) {
      case 0:
	rgb->r = v;
	rgb->g = p3;
	rgb->b = p1;
	return;
      case 1:
	rgb->r = p2;
	rgb->g = v;
	rgb->b = p1;
	return;
      case 2:
	rgb->r = p1;
	rgb->g = v;
	rgb->b = p3;
	return;
      case 3:
	rgb->r = p1;
	rgb->g = p2;
	rgb->b = v;
	return;
      case 4:
	rgb->r = p3;
	rgb->g = p1;
	rgb->b = v;
	return;
      case 5:
	rgb->r = v;
	rgb->g = p1;
	rgb->b = p2;
	return;
    }
    return;
}

/******************** RGB <---> HLS ********************/
void rgb_to_hls(
    rgb_t* rgb,
    double* h,
    double* l,
    double* s
)
{
    double cmin, cmax, cd;

    cmax = MAX3(rgb->r, rgb->g, rgb->b);
    cmin = MIN3(rgb->r, rgb->g, rgb->b);

    *l = (cmax + cmin) / 2;	/* [-1, 1] */
    cd = cmax - cmin;

    if(cd == 0.0)
    {
	*s = 0;
	*h = 0;
	return;
    }

    if(*l <= 0.5)
	*s = cd / (cmax + cmin);
    else
	*s = cd / (2 - cmax - cmin);

    if(rgb->r == cmax)
	*h = (rgb->g - rgb->b) / cd;
    else if(rgb->g == cmax)
	*h = 2 + (rgb->b - rgb->r) / cd;
    else
	*h = 4 + (rgb->r - rgb->g) / cd;

    *h *= 60;
    if(*h < 0)
	*h += 360;
}

void hls_to_rgb(double h, double l, double s, rgb_t* rgb)
{
    double cmin, cmax;

    if (l < 0)
	l = 0;
    if (l > 1)
	l = 1;

    if (s < 0)
	s = 0;
    if (s > 1)
	s = 1;

    h = fmod(h, 360.0);
    if (h < 0)
	h += 360;

    if(l <= 0.5)
    {
	cmin = l * (1 - s);
	cmax = 2 * l - cmin;
    }
    else
    {
	cmax = l * (1 - s) + s;
	cmin = 2 * l - cmax;
    }

    rgb->r = h2v(h + 120.0, cmin, cmax);
    rgb->g = h2v(h        , cmin, cmax);
    rgb->b = h2v(h - 120.0, cmin, cmax);
}

/******************** RGB <---> CMY ********************/
void rgb_to_cmy(
    rgb_t*	rgb,
    double*	c,
    double*	m,
    double*	y)
{
    *c = 1 - rgb->r;
    *m = 1 - rgb->g;
    *y = 1 - rgb->b;
}

void cmy_to_rgb(
    double	c,
    double	m,
    double	y,
    rgb_t*	rgb)
{
    rgb->r = 1 - c;
    rgb->g = 1 - m;
    rgb->b = 1 - y;
}


/******************** RGB <---> CIE ********************/
void rgb_to_cie(rgb_t* rgb, double* x, double* y, double* z)
{
    double r = rgb->r,
	   g = rgb->g,
	   b = rgb->b;

    double X = 0.478 * r + 0.299 * g + 0.175 * b;
    double Y = 0.263 * r + 0.655 * g + 0.081 * b;
    double Z = 0.020 * r + 0.160 * g + 0.908 * b;
    double W = X + Y + Z;

    if(W == 0)
    {
	*x = *y = *z = 0;
	return;
    }

    *x = X / W;
    *y = Y / W;
    *z = Y;
}

void cie_to_rgb(double x, double y, double Y, rgb_t* rgb)
{
    double xYy, xyYy;

    if(y <= 0)
    {
	rgb->r = rgb->g = rgb->b = 1;
	return;
    }

    if(x < 0)
	x = 0;
    if (Y < 0)
	Y = 0;

    xYy = x * Y / y;
    xyYy = (1 - x - y) * Y / y;

    /* CRT */
    rgb->r = 2.739386694386690 * xYy - 1.144708939708940 * Y
	    -0.424074844074844 * xyYy;
    rgb->g =-1.118985713198160 * xYy + 2.028500773973170 * Y
	    +0.033144618976324 * xyYy;
    rgb->b = 0.137976247723133 * xYy - 0.333450588949605 * Y
	    +1.104800777170610 * xyYy;
}

/******************** RGB <---> YIQ ********************/
void rgb_to_yiq(
    rgb_t* rgb,
    double* y,
    double* i,
    double* q
)
{
    double r = rgb->r,
	   g = rgb->g,
	   b = rgb->b;

    *y = 0.299 * r + 0.587 * g + 0.114 * b;
    *i = 0.596 * r - 0.274 * g - 0.322 * b;
    *q = 0.211 * r - 0.522 * g + 0.311 * b;
}

void yiq_to_rgb(
    double y,
    double i,
    double q,
    rgb_t* rgb
)
{
    rgb->r = y + 0.956 * i + 0.623 * q;
    rgb->g = y - 0.272 * i - 0.648 * q;
    rgb->b = y - 1.105 * i + 0.705 * q;
}
