#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>
#include "xmisc.h"
#include "colcnv.h"

#define XCMAP(disp) XDefaultColormap(disp, DefaultScreen(disp))
#define DEFAULT_FONT8  "a14"
#define DEFAULT_FONT16 "k14"

struct font_list_node_t
{
    Display* disp;
    char* fontname;
    XFontStruct* fs;
    struct font_list_node_t* next;
};

static struct font_list_node_t* font_list = NULL;

void DrawRuledVLine(
	Display* disp,
	Drawable drawable,
	GC gc,
	Pixel low_color,
	Pixel high_color,
	int x,
	int y,
	int len)
{
    XSetForeground(disp, gc, low_color);
    XDrawLine(disp, drawable, gc, x, y, x, y + len);

    XSetForeground(disp, gc, high_color);
    XDrawLine(disp, drawable, gc, x+1, y, x+1, y + len);
}

void DrawRuledHLine(
	Display* disp,
	Drawable drawable,
	GC gc,
	Pixel low_color,
	Pixel high_color,
	int x,
	int y,
	int len)
{
    XSetForeground(disp, gc, low_color);
    XDrawLine(disp, drawable, gc, x, y, x + len, y);

    XSetForeground(disp, gc, high_color);
    XDrawLine(disp, drawable, gc, x, y+1, x + len, y+1);
}

Display* OpenDisplay(
	const char* hostname)
{
    Display* disp;

    if((disp = XOpenDisplay(hostname)) == NULL)
    {
	fprintf(stderr, "Can't open display `%s'.\n",
		XDisplayName(hostname));

	if(hostname != NULL)
	    fprintf(stderr, "Check the -display argument");
	else
	    fprintf(stderr, "Check the DISPLAY environment");
	fprintf(stderr, " or `xhost' command in the remote X server.\n");
	exit(1);
    }

    return disp;
}

static double calc_color_diff(int r1, int g1, int b1,
			      int r2, int g2, int b2)
{
    double rd, gd, bd;

    rd = r2 - r1;
    gd = g2 - g1;
    bd = b2 - b1;

    return rd * rd + gd * gd + bd * bd;
}

static long search_near_color(Display* disp, int r, int g, int b)
{
    Visual* vis;
    int vis_class;
    int depth;
    double d, mind;
    int scr;
    static XColor* xc = NULL;
    static int xc_size = 0;
    long i, k;

    scr = DefaultScreen(disp);
    vis = DefaultVisual(disp, scr);
    depth = DefaultDepth(disp, scr);

#if defined(__cplusplus) || defined(c_plusplus)
    vis_class = vis->c_class;	/* C++ class of screen (monochrome, etc.) */
#else
    vis_class = vis->class;	/* class of screen (monochrome, etc.) */
#endif

    if(depth == 1)		/* black or white */
    {
	d = (double)r * r
	  + (double)g * g
	  + (double)b * b;

	if(d > 3.0 * 32768.0 * 32768.0)
	    return (long)WhitePixel(disp, scr);
	return (long)BlackPixel(disp, scr);
    }

    if(vis_class == TrueColor || vis_class == DirectColor)
    {
	return 0;		/* Can't search near color becose maybe too
				 * many colors are allocated.
				 */
    }

    if(xc_size == 0)
    {
	xc_size = DisplayCells(disp, scr);

	if((xc = (XColor *)malloc(sizeof(XColor) * xc_size)) == NULL)
	{
	    fprintf(stderr, "Warning: Can't allocate memory for color cells.\n");
	    xc_size = 0;
	    return 0;
	}
	for(i = 0; i < xc_size; i++)
	    xc[i].pixel = i;
    }

    XQueryColors(disp, DefaultColormap(disp, scr), xc, xc_size);

    mind = calc_color_diff(r, g, b, xc[0].red, xc[0].green, xc[0].blue);

    k = 0;
    for(i = 1; i < xc_size; i++)
    {
	d = calc_color_diff(r, g, b, xc[i].red, xc[i].green, xc[i].blue);
	if(d < mind)
	{
	    mind = d;
	    k = i;
	    if(mind == 0.0)
		break;
	}
    }

#if DEBUG
    printf("color [%04x %04x %04x]->[%04x %04x %04x] (d^2=%f, k=%d)\n",
	   r, g, b,
	   xc[k].red, xc[k].green, xc[k].blue,
	   mind, (int)k);
#endif

    return k;
}

Pixel AllocNamedColor(
	Display* disp,
	const char* color_name)
{
    XColor sdef,		/* screen defined */
           edef;		/* exact defined */
    const char* orig_color_name = color_name;

    /* skip spaces */
    while(*color_name == ' ' || *color_name == '\t')
	color_name++;

    if(*color_name == '#')	/* #rgb or #rrggbb or #rrrrggggbbbb */
    {
	if(!XParseColor(disp, XCMAP(disp), color_name, &edef))
	{
	    fprintf(stderr, "Warning: wrong color specified `%s'.\n",
		    orig_color_name);
	    return 0;
	}
	return AllocRGBColor(disp,
			     (double)edef.red / (double)0xffff,
			     (double)edef.green / (double)0xffff,
			     (double)edef.blue / (double)0xffff);
    }

    if(!XAllocNamedColor(disp, XCMAP(disp), color_name, &sdef, &edef))
    {
	if(!XParseColor(disp, XCMAP(disp), color_name, &edef))
	{
	    fprintf(stderr, "Warning: wrong color specified `%s'.\n",
		    orig_color_name);
	    return 0;
	}
	fprintf(stderr, "Warning: Can't allocate color name `%s'.\n",
		orig_color_name);
	XLookupColor(disp, XCMAP(disp), color_name, &sdef, &edef);
	return search_near_color(disp, sdef.red, sdef.green, sdef.blue);
    }

    return sdef.pixel;
}

Pixel AllocRGBColor(
	Display* disp,
	double red,		/* [0, 1] */
	double green,		/* [0, 1] */
	double blue)		/* [0, 1] */
{
    XColor c;

    c.red = (unsigned short)(red * 0xffff);
    c.green = (unsigned short)(green * 0xffff);
    c.blue = (unsigned short)(blue * 0xffff);

    if(!XAllocColor(disp, XCMAP(disp), &c))
    {
	fprintf(stderr, "Warning: Can't allocate color: "
		"r = %04x, g = %04x, b = %04x\n", c.red, c.green, c.blue);
	return search_near_color(disp, c.red, c.green, c.blue);
    }

    return c.pixel;
}

void PixelToRGB(
	Display* disp,
	Pixel pixel,
	double* r_ret,		/* [0.0, 1.0] */
	double* g_ret,		/* [0.0, 1.0] */
	double* b_ret)		/* [0.0, 1.0] */
{
    XColor c;

    c.pixel = pixel;
    XQueryColor(disp, XCMAP(disp), &c);

    *r_ret = (double)c.red / (double)0xffff;
    *g_ret = (double)c.green / (double)0xffff;
    *b_ret = (double)c.blue / (double)0xffff;
}

void AllocLowHighColor(
	Display* disp,
	Pixel src_color,
	Pixel* low_color_ret,
	Pixel* high_color_ret)
{
    double r0, g0, b0;
    rgb_t rgb;
    double h, l, s;
    static const double d = 0.3;

    PixelToRGB(disp, src_color, &r0, &g0, &b0);

    rgb.r = r0;
    rgb.g = g0;
    rgb.b = b0;
    rgb_to_hls(&rgb, &h, &l, &s);
    l = MAX(l - d, 0.0);
    hls_to_rgb(h, l, s, &rgb);
    *low_color_ret = AllocRGBColor(disp, rgb.r, rgb.g, rgb.b);

    rgb.r = r0;
    rgb.g = g0;
    rgb.b = b0;
    rgb_to_hls(&rgb, &h, &l, &s);
    l = MIN(l + d, 1.0);
    hls_to_rgb(h, l, s, &rgb);
    *high_color_ret = AllocRGBColor(disp, rgb.r, rgb.g, rgb.b);
}

void DrawBorder(
	Display* disp,
	Drawable drawable,
	GC gc,
	int x,
	int y,
	int width,
	int height,
        int bd_width,
	Pixel low_color,
	Pixel high_color,
	Bool status)		/* True if it looks like convex,
				   otherwise False (concave). */
{
    int i;
    Pixel upper_left;
    Pixel bottom_right;

    if(status)			/* convex */
    {
	upper_left = high_color;
	bottom_right = low_color;
    }
    else
    {
	upper_left = low_color;
	bottom_right = high_color;
    }

    XSetForeground(disp, gc, upper_left);
    for(i = 0; i < bd_width; i++)
    {
        XDrawLine(disp, drawable, gc,
		  x + i, y + i, x + i, height - i + y - 1);
        XDrawLine(disp, drawable, gc,
		  x + i, y + i, x + width - i - 1, y + i);
    }

    XSetForeground(disp, gc, bottom_right);
    for(i = 0; i < bd_width; i++)
    {
        XDrawLine(disp, drawable, gc,
		  x + width - i - 1, y + height - i - 1,
		  x + i, y + height - i - 1);
        XDrawLine(disp, drawable, gc,
		  x + width - i - 1, y + height - i - 1,
		  x + width - i - 1, y + i);
    }
}

void MapWindow(
	Display* disp,
	Window win,
	Bool   iswait)		/* is waiting until window is maped */
{
    XWindowAttributes at;
    XEvent e;

    if(iswait == False)
    {
        XMapWindow(disp, win);
        return;
    }

    XGetWindowAttributes(disp, win, &at);

    if(at.map_state == IsViewable)
	return;

    if(at.backing_store == WhenMapped ||
       at.backing_store == Always)
    {
        XMapWindow(disp, win);
        return;
    }

    XSelectInput(disp, win, ExposureMask);
    XMapWindow(disp, win);
    XFlush(disp);

    while(XNextEvent(disp, &e), e.type != Expose)
        ;

    /* recover the original event mask */
    XSelectInput(disp, win, at.your_event_mask);
}

Window CreateWindow(
	Display* disp,
	Window parent,		/* None is RootWindow */
	int x,
	int y,
	int width,
	int height,
	Pixel bg)		/* background color */
{
    Window win;

    if(parent == None)
	parent = DefaultRootWindow(disp);
    win = XCreateSimpleWindow(disp, parent, x, y, width, height,
			      0, 0, bg);

    return win;
}

void GetWindowSize(
	Display* disp,
	Drawable drawable,
	int* width_ret,
	int* height_ret)
{
    unsigned int ui;
    int i;
    Window w;

    XGetGeometry(disp, drawable, &w, &i, &i,
		 (unsigned int *)width_ret, (unsigned int *)height_ret,
		 &ui, &ui);
}

Pixmap CreatePixmap(
	Display* disp,
	Window parent,		/* None is RootWindow */
	int width,
	int height,
	Pixel bg)		/* background color */
{
    Pixmap pm;
    static GC gc = NULL;
    static Display* gcdisp = NULL;

    if(parent == None)
	parent = DefaultRootWindow(disp);
    pm = XCreatePixmap(disp, parent, width, height,
		       DefaultDepth(disp, DefaultScreen(disp)));

    if(gc == NULL)
    {
	gc = XCreateGC(disp, DefaultRootWindow(disp), 0, NULL);
	gcdisp = disp;
    }
    else if(disp != gcdisp)
    {
	XFreeGC(gcdisp, gc);
	gc = XCreateGC(disp, DefaultRootWindow(disp), 0, NULL);
	gcdisp = disp;
    }

    XSetForeground(disp, gc, bg);
    XFillRectangle(disp, pm, gc, 0, 0, width, height);

    return pm;
}

void SetWindowBorder(
	Display* disp,
	Window win,
	GC gc,
	int width,		/* border width */
	Pixel bg,
	Pixel low_color,
	Pixel high_color,
	Bool status)		/* True if it looks like convex,
				   otherwise False (concave). */
{
    int w, h;
    Pixmap pm;

    GetWindowSize(disp, win, &w, &h);
    pm = CreatePixmap(disp, win, w, h, bg);

    DrawBorder(disp, pm, gc, 0, 0, w, h, width, low_color, high_color, status);

    XSetWindowBackgroundPixmap(disp, win, pm);
    XFreePixmap(disp, pm);
}

Display* CheckMultiDisplayEvent(
	Display** disp_list,
	int ndisp,
	struct timeval* tout)
{
    static int nfd;
    static int* fds;
    int i, width, rval;
    fd_set readfds;

    for(i = 0; i < ndisp; i++)
    {
	XFlush(disp_list[i]);
	if(QLength(disp_list[i]))
	    return disp_list[i];
    }

    if(tout != NULL &&
       tout->tv_sec == 0 &&
       tout->tv_usec == 0)
	return NULL;

    if(fds == NULL)
	fds = (int *)malloc(sizeof(int) * ndisp);
    else if(nfd < ndisp)
    {
	free(fds);
	fds = (int *)malloc(sizeof(int) * ndisp);
    }

    width = 0;
    for(i = 0; i < ndisp; i++)
	fds[i] = ConnectionNumber(disp_list[i]);

    FD_ZERO(&readfds);
    for(i = 0; i < ndisp; i++)
	FD_SET(fds[i], &readfds);

    for(i = width = 0; i < ndisp; i++)
	if(fds[i] > width)
	    width = fds[i];
    width++;

    rval = select(width, &readfds, NULL, NULL, tout);
    if(rval < 0)
    {
	perror("select");
	sleep(1);
	return NULL;
    }

    if(rval == 0)
	return NULL;

    for(i = 0; i < ndisp; i++)
	if(FD_ISSET(fds[i], &readfds) != 0)
	    return disp_list[i];
    return NULL;
}

static struct font_list_node_t* new_font_node(Display* disp,
					      const char* fontname)
{
    struct font_list_node_t* p;

#ifdef DEBUG
    printf("Loading font `%s'\n", fontname);
#endif

    if((p = (struct font_list_node_t *)
	malloc(sizeof(struct font_list_node_t))) == NULL)
    {
	perror("malloc");
	return NULL;
    }

    if((p->fontname = strdup(fontname)) == NULL)
    {
	perror("strdup");
	free(p);
	return NULL;
    }

    if((p->fs = XLoadQueryFont(disp, fontname)) == NULL)
    {
	fprintf(stderr, "Can't find font `%s'\n", fontname);
	free(p);
	free(p->fontname);
	return NULL;
    }

    p->disp = disp;
    p->next = NULL;
    return p;
}

XFontStruct* LoadFont(
	Display* disp,
	const char* fontname)
{
    struct font_list_node_t* p = font_list;

    if(p == NULL)
    {
	if((font_list = new_font_node(disp, fontname)) == NULL)
	    return NULL;
	return font_list->fs;
    }

    for(; p->next != NULL; p = p->next)
	if(p->disp == disp && !strcmp(p->fontname, fontname))
	    return p->fs;
    if(p->disp == disp && !strcmp(p->fontname, fontname))
	return p->fs;
    if((p->next = new_font_node(disp, fontname)) == NULL)
	return NULL;
    return p->next->fs;
}

Bool IsKeyPressed(
	Display* disp,
	KeySym sym)
{
    char key_vector[32];
    KeyCode code;
    int keybit;
    unsigned int byteindex;	/* keycode's byte */
    unsigned int bitindex;	/* keycode's bit */

    code = XKeysymToKeycode(disp, sym);
    if(code == NoSymbol)
	return 0;

    byteindex = (code / 8);
    bitindex = code & 7;
    XQueryKeymap(disp, key_vector);
    keybit = (1 & (key_vector[byteindex] >> bitindex));

    return keybit;
}

void CompressEvent(
	Display* disp,
	Window win,
	int event_type,
	XEvent* e)
{
    do
    {
/*	XFlush(disp); */
	XSync(disp, False);
    } while(XCheckTypedWindowEvent(disp, win, event_type, e));
}

void CompressMotionEvent(
	Display* disp,
	Window win,
	XEvent* e)
{
    CompressEvent(disp, win, MotionNotify, e);
}

void CompressKeyPressEvent(
	Display* disp,
	Window win,
	XEvent* e)
{
    CompressEvent(disp, win, KeyPress, e);
}

void QueryPointer(
	Display* disp,
	Window win,
	int* x_ret,
	int* y_ret)
{
    Window wdmy;
    int idmy;
    unsigned int udmy;

    if(x_ret == NULL)
      x_ret = &idmy;
    if(y_ret == NULL)
      y_ret = &idmy;

    if(!XQueryPointer(disp, win,
		      &wdmy, &wdmy, 
		      x_ret, y_ret, &idmy, &idmy, &udmy))
    {
	fprintf(stderr, "Warning: Could not query pointer.\n");
	*x_ret = *y_ret = 0;
	return;
    }
}

void QueryRootPointer(
	Display* disp,
	int* x_ret,
	int* y_ret)
{
    Window wdmy;
    int idmy;
    unsigned int udmy;

    if(!XQueryPointer(disp, DefaultRootWindow(disp),
		      &wdmy, &wdmy,
		      &idmy, &idmy, x_ret, y_ret, &udmy))
    {
	fprintf(stderr, "Warning: Could not query pointer.\n");
	*x_ret = *y_ret = 0;
	return;
    }
}

void SetFixWindowSize(
	Display* disp,
	Window win,
	int width,
	int height)
{
    XSizeHints hints;

    hints.flags      = PMinSize | PMaxSize;
    hints.min_width  = hints.max_width  = width;
    hints.min_height = hints.max_height = height;

    XSetNormalHints(disp, win, &hints);
}

CharFont* CreateCharFont(
	Display* disp,
	const char* fontname8,
	const char* fontname16)
{
    CharFont* font;

    if((font = (CharFont *)malloc(sizeof(CharFont))) == NULL)
    {
	perror("malloc");
	return NULL;
    }
    font->gc8  = XCreateGC(disp, DefaultRootWindow(disp), 0, NULL);
    font->gc16 = XCreateGC(disp, DefaultRootWindow(disp), 0, NULL);

    if(fontname8)
    {
	if((font->font8 = LoadFont(disp, fontname8)) == NULL)
	    font->font8 = LoadFont(disp, DEFAULT_FONT8);
    }
    else
	font->font8 = LoadFont(disp, DEFAULT_FONT8);

    if(fontname16)
    {
	if((font->font16 = LoadFont(disp, fontname16)) == NULL)
	    font->font16 = LoadFont(disp, DEFAULT_FONT16);
    }
    else
	font->font16 = LoadFont(disp, DEFAULT_FONT16);

    if(font->font8 == NULL || font->font16 == NULL)
	exit(1);

    XSetFont(disp, font->gc8,  font->font8->fid);
    XSetFont(disp, font->gc16, font->font16->fid);

    return font;
}

int TextWidth(
	CharFont* font,
	const char* text)
{
    return TextWidthN(font, text, strlen(text));
}

int TextWidthN(
	CharFont* font,
	const char* text,
	int n)
{
    int width;
    XChar2b xc;
    XFontStruct* fs1 = font->font8;
    XFontStruct* fs2 = font->font16;
    const char* limit;

    width = 0;
    limit = text + n;
    while (text < limit)
    {
	if(*text & 0x80)
	{
	    xc.byte1 = text[0] & 0x7f;
	    xc.byte2 = text[1] & 0x7f;
	    width += XTextWidth16(fs2, &xc, 1);
	    text += 2;
	}
	else
	{
	    width += XTextWidth(fs1, text, 1);
	    text++;
	}
    }
    return width;
}

void DrawStringN(
	Display* disp,
	Drawable win,
	CharFont* font,
	Pixel fg,
	int x, int y,
	int width, int height,
	const char* string,
	int n,
	int alignment)
{
    int text_width, text_height;
    int basex, basey;
    int dy1, dy2;
    XFontStruct* fs1 = font->font8;
    XFontStruct* fs2 = font->font16;
    XChar2b xc;
    const char* limit;

    if(!*string)
	return;

    text_height = TextHeight(font);
    text_width  = TextWidthN(font, string, n);

    switch(alignment)
    {
      case Center:
	basex = x + (width - text_width) / 2;
	basey = y + (height - text_height) / 2 + text_height;
	break;
      case Noth:
	basex = x + (width - text_width) / 2;
	basey = y + text_height;
	break;
      case NothEast:
	basex = x + width - text_width;
	basey = y + text_height;
	break;
      case East:
	basex = x + width - text_width;
	basey = y + (height - text_height) / 2 + text_height;
	break;
      case SouthEast:
	basex = x + width - text_width;
	basey = y + text_height;
	break;
      case South:
	basex = x + (width - text_width) / 2;
	basey = y + text_height;
	break;
      case SouthWest:
	basex = x;
	basey = y + text_height;
	break;
      case West:
	basex = x;
	basey = y + (height - text_height) / 2 + text_height;
	break;
      case NothWest:
	basex = x;
	basey = y + text_height;
	break;
      default:
	basex = x;
	basey = y + text_height;
    }

    if(*string & 0x80)
	basex -= fs2->min_bounds.lbearing;
    else
	basex -= fs1->min_bounds.lbearing;

    dy1 = -fs1->max_bounds.descent;
    dy2 = -fs2->max_bounds.descent;

    x = basex;
    XSetForeground(disp, font->gc8,  fg);
    XSetForeground(disp, font->gc16, fg);

    limit = string + n;
    while(string < limit)
    {
	if(*string & 0x80)
	{
	    xc.byte1 = string[0] & 0x7f;
            xc.byte2 = string[1] & 0x7f;
	    XDrawString16(disp, win, font->gc16, x, basey + dy2, &xc, 1);
	    x += XTextWidth16(fs2, &xc, 1);
	    string += 2;
	}
	else
	{
	    XDrawString(disp, win, font->gc8, x, basey + dy1, string, 1);
	    x += XTextWidth(fs1, string, 1);
	    string++;
	}
    }
}

void DrawString(
	Display* disp,
	Drawable win,
	CharFont* font,
	Pixel fg,
	int x, int y,
	int width, int height,
	const char* string,
	int alignment)
{
    DrawStringN(disp, win, font, fg, x, y, width, height,
		string, strlen(string), alignment);
}

static const char* skip_line(const char* s)
{
    while(*s++ != '\n')
	;
    return s;
}

Pixmap CreateIconImagePixmap(
	Display* disp,
	GC gc,
	Pixel* icon_color_map,
	const char* icon_image,
	int* width_return,
	int* height_return)
{
    int width, height, x, y;
    Pixmap pm;
    Pixel p, pixel;
    char ch;

#if 0
    sscanf(icon_image, "%d %d%[^\n]", &width, &height); /* warning: too few arguments for format ?? (reported on gcc-2.7.2 with -Wall flag) */
#endif
    sscanf(icon_image, "%d %d", &width, &height);
    icon_image = skip_line(icon_image);

    p = icon_color_map['0'];
    pm = CreatePixmap(disp, None, width, height, p);
    XSetForeground(disp, gc, p);

    for(y = 0; y < height; y++)
    {
	for(x = 0; x < width; x++)
	{
	    ch = icon_image[(width + 1) * y + x];
	    if(ch < 0)
		continue;
	    pixel = icon_color_map[ch];
	    if(p != pixel)
	    {
		p = pixel;
		XSetForeground(disp, gc, p);
	    }
	    XDrawPoint(disp, pm, gc, x, y);
	}
    }

    if(width_return != NULL)
	*width_return = width;
    if(height_return != NULL)
	*height_return = height;
    return pm;
}

unsigned long GetEventMask(
	Display* disp,
	Window win)
{
    XWindowAttributes at;
    XGetWindowAttributes(disp, win, &at);
    return at.your_event_mask;
}

Window ParentWindow(
	Display* disp,
	Window win)
{
    Window parent, wdmy;
    Window* cld = NULL;
    unsigned int n;

    XQueryTree(disp, win, &wdmy, &parent, &cld, &n);
    if(cld != NULL)
	XFree((void *)cld);
    return parent;
}

void QueryWindowPositionFromRootWindow(
	Display* disp,
	Window win,
	int* x_ret,
	int* y_ret)
{
    Window root = DefaultRootWindow(disp);
    int x, y;

    x = y = 0;
    while(win != root)
    {
	int rx, ry;
	Window wdmy;
	unsigned int idmy, bd;

	XGetGeometry(disp, win, &wdmy, &rx, &ry,
		     &idmy, &idmy, &bd, &idmy);
	win = ParentWindow(disp, win);
	x += rx + bd;
	y += ry + bd;
    }

    *x_ret = x;
    *y_ret = y;
}

void DrawChecker(
	Display* disp,
	Drawable drawable,
	GC gc,
	int x,
	int y,
	int width,
	int height)
{
    int tx, ty;

    /* FIXME */
    for(ty = 0; ty < height; ty++)
	for(tx = ty % 2; tx < width; tx += 2)
	    XDrawPoint(disp, drawable, gc, x + tx, y + ty);
}

XImage* GetImageData(Display* disp, Drawable da, int x, int y,
		     int width, int height)
{
    return XGetImage(disp, da, x, y, width, height, AllPlanes, ZPixmap);
}

#define SQRT3D2 0.86602540378444
void DrawTriangle(Display* disp, Window win, GC gc,
		  int x, int y, int bd, double r, double th,
		  Pixel bg, Pixel lbg, Pixel hbg, Bool isconvex)
{
    int i;
    XPoint point[4];
    double r1, r2;
    double tx1, ty1, tx, ty;
    Pixel c;

    if(isconvex)
    {
	c = hbg;
	hbg = lbg;
	lbg = c;
    }

    r1 = r;
    r2 = r - 2 * bd;

    tx1 = cos(th);
    ty1 = sin(th);

    for(i = 0; i < 3; i++)
    {
	point[i].x = (short)(x + 0.5 + r2 * tx1);
	point[i].y = (short)(y - 0.5 - r2 * ty1);
	tx = tx1;
	ty = ty1;
	tx1 = -0.5 * tx - SQRT3D2 * ty;
	ty1 = SQRT3D2 * tx - 0.5 * ty;
    }
    XSetForeground(disp, gc, bg);
    XFillPolygon(disp, win, gc, point, 3, Convex, CoordModeOrigin);

    point[3].x = (short)(x + 0.5 + r1 * tx1);
    point[3].y = (short)(y - 0.5 - r1 * ty1);
    point[2].x = (short)(x + 0.5 + r2 * tx1);
    point[2].y = (short)(y - 0.5 - r2 * ty1);
    th = fmod(th, 2 * M_PI);
    if(th < 0)
	th += 2 * M_PI;
    for(i = 0; i < 3; i++)
    {
	point[0] = point[3];
	point[1] = point[2];

	if(th >= M_PI / 4 - M_PI / 3 && th <  5 * M_PI / 4 - M_PI / 3)
	    c = lbg;
	else
	    c = hbg;

	th += 2 * M_PI / 3;
	if(th >= 2 * M_PI)
	    th -= 2 * M_PI;

	tx = tx1;
	ty = ty1;
	tx1 = -0.5 * tx - SQRT3D2 * ty;
	ty1 = SQRT3D2 * tx - 0.5 * ty;

	point[3].x = (short)(x + 0.5 + r1 * tx1);
	point[3].y = (short)(y - 0.5 - r1 * ty1);
	point[2].x = (short)(x + 0.5 + r2 * tx1);
	point[2].y = (short)(y - 0.5 - r2 * ty1);

	XSetForeground(disp, gc, c);
	XFillPolygon(disp, win, gc, point, 4, Convex, CoordModeOrigin);
    }
}

#define XMISC_PROP_NAME "__XMISC_PROP"
static Atom xmisc_prop = None;

unsigned long GetPropertyDataSize(Display* disp, Atom type)
{
    Atom ret_type;
    int format;
    unsigned long nitems, after;
    unsigned char* data;

    if(xmisc_prop == None)
	xmisc_prop = XInternAtom(disp, XMISC_PROP_NAME, False);

    data = NULL;
    XGetWindowProperty(disp, DefaultRootWindow(disp), xmisc_prop, 0, 0,
		       False, type, &ret_type,
		       &format, &nitems, &after, &data);
    if(data != NULL)
	XFree(data);
    
    return after;
}

char* GetPropertyValue(Display* disp, Atom type, unsigned long* nbytes)
{
    unsigned char* buff;
    char* data;
    unsigned long n;
    Atom ret_type;
    int format;
    unsigned long nitems, after;

    *nbytes = GetPropertyDataSize(disp, type);
    if((data = (char *)malloc(*nbytes)) == NULL)
	return NULL;

    n = 0;
    do
    {
	buff = NULL;
	XGetWindowProperty(disp, DefaultRootWindow(disp), xmisc_prop,
			   (long)n/4, MAX_PROP_CHUNK_SIZE,
			   False, type, &ret_type,
			   &format, &nitems, &after, &buff);
	if(buff == NULL)
	    break;
	memcpy(data + n, buff, nitems);
	XFree(buff);
	n += nitems;
    } while(after > 0 && n < *nbytes);
    *nbytes = n;
    return data;
}

void AppendPropertyValue(Display* disp, Atom type, void* value, int nbytes)
{
    unsigned char* data;

    if(xmisc_prop == None)
	xmisc_prop = XInternAtom(disp, XMISC_PROP_NAME, False);

    data = (unsigned char *)value;
    while(MAX_PROP_CHUNK_SIZE < nbytes)
    {
	AppendPropertyValue(disp, type, data, MAX_PROP_CHUNK_SIZE);
	nbytes -= MAX_PROP_CHUNK_SIZE;
	data += MAX_PROP_CHUNK_SIZE;
    }
    XChangeProperty(disp, DefaultRootWindow(disp),
		    xmisc_prop, type, 8, PropModeAppend, data, nbytes);
}

void SetPropertyValue(Display* disp, Atom type, void* value,
		      unsigned long nbytes)
{
    unsigned char* data;
    int n;

    if(xmisc_prop == None)
	xmisc_prop = XInternAtom(disp, XMISC_PROP_NAME, False);

    data = (unsigned char *)value;
    n = (int)MIN(nbytes, MAX_PROP_CHUNK_SIZE);
    XChangeProperty(disp, DefaultRootWindow(disp),
		    xmisc_prop, type, 8, PropModeReplace, data, n);
    if(nbytes > MAX_PROP_CHUNK_SIZE)
	AppendPropertyValue(disp, type,
		    data + MAX_PROP_CHUNK_SIZE,
			    (int)(nbytes - MAX_PROP_CHUNK_SIZE));
}

const char* EventTypeString(int type)
{
    switch(type)
    {
      case KeyPress:		return "KeyPress";
      case KeyRelease:		return "KeyRelease";
      case ButtonPress:		return "ButtonPress";
      case ButtonRelease:	return "ButtonRelease";
      case MotionNotify:	return "MotionNotify";
      case EnterNotify:		return "EnterNotify";
      case LeaveNotify:		return "LeaveNotify";
      case FocusIn:		return "FocusIn";
      case FocusOut:		return "FocusOut";
      case KeymapNotify:	return "KeymapNotify";
      case Expose:		return "Expose";
      case GraphicsExpose:	return "GraphicsExpose";
      case NoExpose:		return "NoExpose";
      case VisibilityNotify:	return "VisibilityNotify";
      case CreateNotify:	return "CreateNotify";
      case DestroyNotify:	return "DestroyNotify";
      case UnmapNotify:		return "UnmapNotify";
      case MapNotify:		return "MapNotify";
      case MapRequest:		return "MapRequest";
      case ReparentNotify:	return "ReparentNotify";
      case ConfigureNotify:	return "ConfigureNotify";
      case ConfigureRequest:	return "ConfigureRequest";
      case GravityNotify:	return "GravityNotify";
      case ResizeRequest:	return "ResizeRequest";
      case CirculateNotify:	return "CirculateNotify";
      case CirculateRequest:	return "CirculateRequest";
      case PropertyNotify:	return "PropertyNotify";
      case SelectionClear:	return "SelectionClear";
      case SelectionRequest:	return "SelectionRequest";
      case SelectionNotify:	return "SelectionNotify";
      case ColormapNotify:	return "ColormapNotify";
      case ClientMessage:	return "ClientMessage";
      case MappingNotify:	return "MappingNotify";
    }
    return "Unknown event type";
}
