#include <stdio.h>
#include <string.h>
#include "ssdefs.h"
#include "audriv.h"
#include "aubuff.h"
#include "bucket.h"
#include "misc.h"
#include "usffd.h"

static AudioBuffer* head_audio_buffer = NULL; /* start of chain list */

#if 0	/* soundscope_interval_hook */
static Bool default_soundscope_interval_hook(int nread, int n)
{
    return True;
}
Bool (* soundscope_interval_hook)(int nread, int n) = default_soundscope_interval_hook;
#endif

void initialize_audio_buffer(void)
{
    make_new_audio_buffer(AUDIO_DEFAULT_BUFF_NAME);
    make_new_audio_buffer(AUDIO_COPY_BUFF_NAME);
    make_new_audio_buffer(AUDIO_TEMP_BUFF_NAME);
    clear_audio_buffer_mark();
}

AudioBuffer* get_audio_buffer(const char* name)
{
    AudioBuffer* tail;

    for(tail = head_audio_buffer; tail; tail = tail->next)
	if(tail->name &&
	   !strcmp(name, tail->name))
	    break;
    return tail;
}

AudioBuffer* make_new_audio_buffer(const char* name)
{
    AudioBuffer* buf;
    char* buf_name;
    int cnt;

    if((buf = (AudioBuffer *)xmalloc(sizeof(AudioBuffer))) == NULL)
	return NULL;

    if((buf_name = (char *)xmalloc(strlen(name) + 16)) == NULL)
    {
	xfree(buf);
	return NULL;
    }

    strcpy(buf_name, name);
    for(cnt = 2; get_audio_buffer(buf_name) != NULL; cnt++)
	sprintf(buf_name, "%s<%d>", name, cnt);

    buf->name = buf_name;
    buf->filename = NULL;
    buf->buffer = NULL;
    buf->nsamples = 0;
    buf->nchannels = 1;
    buf->sample_rate = 8000;
    buf->file_encoding = USFFD_G711_ULAW;
    buf->hdr_type = -1;
    buf->record_flag = False;
    buf->mark = False;
    buf->next = head_audio_buffer;
    buf->alloc_samples = 0;
    head_audio_buffer = buf;

    return buf;
}

void erase_audio_buffer(AudioBuffer* buf)
{
    int i;

    if(is_audio_buffer_empty(buf))
	return;

    if(buf->buffer != NULL)
    {
	for(i = 0; i < buf->nchannels; i++)
	    xfree(buf->buffer[i]);
	xfree(buf->buffer);
	buf->buffer = NULL;
	buf->alloc_samples = 0;
    }
    buf->nsamples = 0;
    buf->nchannels = 0;
}

Bool audio_buffer_make_buffer(AudioBuffer* b, long nsamples, int nchannels)
{
    int i;

    if(b->buffer != NULL)
    {
	if(b->alloc_samples >= nsamples && b->nchannels == nchannels)
	{
	    b->nsamples = nsamples;
	    return True;
	}
	erase_audio_buffer(b);
    }
    b->alloc_samples = nsamples;
    b->nchannels = nchannels;
    b->nsamples = nsamples;
    
    if((b->buffer = (short **)xmalloc2(b->nchannels, 0)) == NULL)
    {
	kill_audio_buffer(b);
	return False;
    }

    if(nsamples > 0)
    {
	for(i = 0; i < b->nchannels; i++)
	{
	    if((b->buffer[i] = (short *)xmalloc(b->nsamples * sizeof(short)))
	       == NULL)
	    {
		kill_audio_buffer(b);
		return False;
	    }
	}
    }

    return True;
}

static Bool audio_buffer_extend_buffer(AudioBuffer* b, long nsamples)
{
    int nch = b->nchannels;
    int i;
    long more;
    short** data;

    if(b->nsamples >= nsamples)
	return True;
    more = nsamples - b->nsamples; /* > 0 */
    if(b->alloc_samples >= nsamples)
    {
	for(i = 0; i < nch; i++)
	    memset(b->buffer[i] + b->nsamples, 0, more * sizeof(short));
	b->nsamples = nsamples;
	return True;
    }

    if((data = (short **)xmalloc2(nch, nsamples * sizeof(short))) == NULL)
	return False;

    for(i = 0; i < nch; i++)
    {
	memcpy(data[i], b->buffer[i], b->nsamples * sizeof(short));
	memset(data[i] + b->nsamples, 0, more * sizeof(short));
    }
    xfree2(b->buffer, nch);
    b->buffer = data;
    b->nsamples = nsamples;
    b->alloc_samples = nsamples;

    return True;
}

static void kill_audio_buffer1(AudioBuffer* buf)
{
    if(buf == NULL)
	return;
    erase_audio_buffer(buf);
    xfree(buf->name);
    xfree(buf->filename);
    buf->name = buf->filename = NULL;
}

void kill_audio_buffer(AudioBuffer* buf)
{
    AudioBuffer* cur;
    AudioBuffer* prev;

    kill_audio_buffer1(buf);

    /* head_audio_buffer  buf  */
    prev = NULL;
    cur = head_audio_buffer;
    while(cur)
    {
	if(cur == buf)
	    break;
	prev = cur;
	cur = cur->next;
    }
    if(cur != NULL)		/* found */
    {
	if(prev != NULL)
	    prev->next = cur->next;
	else
	    head_audio_buffer = head_audio_buffer->next;
    }

    xfree(buf);
}

AudioBuffer* make_usffd_audio_buffer(const USFFD* usffd, const char* name)
{
    AudioBuffer* b;
    long n, ch;

    if((b = make_new_audio_buffer(name)) == NULL)
	return NULL;
    if((b->filename = xstrdup(usffd->filename)) == NULL)
    {
	kill_audio_buffer(b);
	return NULL;
    }
    b->sample_rate = usffd->sample_rate;
    b->file_encoding = usffd->data_encoding;
    b->hdr_type = usffd->hdr_type;

    ch = usffd->nchannels;
    n = usffd->data_size /
	(usffd_encoding_sample_size[usffd->data_encoding] * ch);
	 
    if(audio_buffer_make_buffer(b, n, ch) == False)
	return NULL;
    return b;
}

void clear_audio_buffer_mark(void)
{
    AudioBuffer* p;

    for(p = head_audio_buffer; p; p = p->next)
	p->mark = False;
}

void audio_buffer_garbage_collect(void)
{
    AudioBuffer* head;
    AudioBuffer* p;

    head = NULL;
    while(head_audio_buffer)
    {
	p = head_audio_buffer;
	head_audio_buffer = head_audio_buffer->next;

	if(p->mark || (p->name != NULL && p->name[0] == ' '))
	{
	    p->next = head;
	    head = p;
	}
	else
	    kill_audio_buffer1(p);
    }

    head_audio_buffer = head;
    if(head_audio_buffer == NULL)
	make_new_audio_buffer(AUDIO_DEFAULT_BUFF_NAME);
}


/*
 * audio editing
 */

Bool audio_buffer_cut_region(AudioBuffer* aubuff,
			     long start, long end)
{
    if(audio_buffer_copy_region(aubuff, start, end) == False)
	return False;
    return audio_buffer_delete_region(aubuff, start, end);
}

Bool audio_buffer_copy_region(AudioBuffer* aubuff, long start, long end)
{
    long n;
    int nch, i;
    AudioBuffer* copy;

    nch = aubuff->nchannels;
    if(start > end)
    {
	long tmp;
	SWAP(start, end, tmp);
    }
    FIXREGION(start, 0, aubuff->nsamples);
    FIXREGION(end,   0, aubuff->nsamples);
    n = end - start;

    copy = get_audio_buffer(AUDIO_COPY_BUFF_NAME);
    if(audio_buffer_make_buffer(copy, n, nch) == False)
	return False;
    for(i = 0; i < nch; i++)
	memcpy(copy->buffer[i], aubuff->buffer[i] + start, n * sizeof(short));
    copy->sample_rate = aubuff->sample_rate;

    return True;
}

Bool audio_buffer_delete_region(AudioBuffer* aubuff,
				long start, long end)
{
    long n;
    int nch;
    short* data;
    int i;

    if(start > end)
    {
	long tmp;
	SWAP(start, end, tmp);
    }

    FIXREGION(start, 0, aubuff->nsamples);
    FIXREGION(end,   0, aubuff->nsamples);

    if(start == end)
	return True;
    n = aubuff->nsamples - end;
    nch = aubuff->nchannels;

    for(i = 0; i < nch; i++)
    {
	data = aubuff->buffer[i];
	memmove(data + start, data + end, n * sizeof(short));
    }

    aubuff->nsamples -= end - start;

    return True;
}

Bool convert_audio_format(AudioBuffer* from_buff, /* no modified */
			  AudioBuffer* to_buff,	  /* erase the original */
			  long start, long end,
			  long nchannels, long sample_rate)
{
    AudioBucket *bucket[MAX_BUCKET_NCHANNELS], *b, *p;
    long i, j, n, to_buff_size;

    FIXREGION(start, 0, from_buff->nsamples);
    FIXREGION(end, 0, from_buff->nsamples);
    n = end - start;
    erase_audio_buffer(to_buff);
    if(n == 0)
	return True;
    if(from_buff->nchannels == nchannels &&
       from_buff->sample_rate == sample_rate)
    {
	if(audio_buffer_make_buffer(to_buff, n, nchannels) == False)
	    return False;
	to_buff->sample_rate = sample_rate;
	for(i = 0; i < from_buff->nchannels; i++)
	    memcpy(to_buff->buffer[i], from_buff->buffer[i] + start,
		  n * sizeof(short));
	return True;
    }

    for(i = 0; i < from_buff->nchannels; i++)
    {
	bucket[i] = buffer_to_bucket_list(from_buff->buffer[i] + start, n);
	bucket[i] = convert_sample_rate(bucket[i],
					from_buff->sample_rate,
					sample_rate);
    }
    b = mix_audio_channel(bucket, from_buff->nchannels, nchannels);

    to_buff_size = 0;
    for(p = b; p; p = p->next)
	to_buff_size += p->len / nchannels;
    if(audio_buffer_make_buffer(to_buff, to_buff_size, nchannels) == False)
	return False;

    n = 0;
    for(p = b; p; p = p->next)
    {
	long jn     = p->len / nchannels;
	short* from = p->data;
	short** to  = to_buff->buffer;

	for(j = 0; j < jn; j++)
	    for(i = 0; i < nchannels; i++)
		to[i][n + j] = *from++;
	n += jn;
    }

    return True;
}

long insert_audio_buffer(AudioBuffer* from_buff, AudioBuffer* to_buff,
			 long to_pos)
{
    short** data;
    int nch, sr, i;
    long nsamples;

    if(is_audio_buffer_empty(from_buff))
	return 0;		/* from buffer is empty */

    FIXREGION(to_pos, 0, to_buff->nsamples);
    if(is_audio_buffer_empty(to_buff))
    {
	/* to buffer is empty */
	nsamples = from_buff->nsamples;
	if(audio_buffer_make_buffer(to_buff, nsamples,
				    from_buff->nchannels) == False)
	    return -1;
	to_buff->sample_rate = from_buff->sample_rate;
	nch = from_buff->nchannels;
	for(i = 0; i < nch; i++)
	    memcpy(to_buff->buffer[i], from_buff->buffer[i],
		   nsamples * sizeof(short));
	return nsamples;
    }

    nch = to_buff->nchannels;
    sr  = to_buff->sample_rate;

    if(from_buff->nchannels != nch || from_buff->sample_rate != sr)
    {
	AudioBuffer* temp;

	temp = get_audio_buffer(AUDIO_TEMP_BUFF_NAME);
	if(convert_audio_format(from_buff, temp,
				0, from_buff->nsamples,
				nch, sr) == False)
	    return -1;
	from_buff = temp;
    }

    nsamples = to_buff->nsamples + from_buff->nsamples;

    if(nsamples > to_buff->alloc_samples)
    {
	if((data = (short **)xmalloc2(nch, nsamples * sizeof(short)))
	   == NULL)
	    return -1;
	for(i = 0; i < nch; i++)
	{
	    memcpy(data[i], to_buff->buffer[i], to_pos * sizeof(short));
	    memcpy(data[i] + to_pos, from_buff->buffer[i],
		   from_buff->nsamples * sizeof(short));
	    memcpy(data[i] + to_pos + from_buff->nsamples,
		   to_buff->buffer[i] + to_pos,
		   (to_buff->nsamples - to_pos) * sizeof(short));
	}
	xfree2(to_buff->buffer, nch);
	to_buff->buffer = data;
	to_buff->alloc_samples = nsamples;
    }
    else
    {
	data = to_buff->buffer;
	for(i = 0; i < nch; i++)
	{
	    memmove(data[i] + to_pos + from_buff->nsamples,
		    data[i] + to_pos,
		    (to_buff->nsamples - to_pos) * sizeof(short));
	    memcpy(data[i] + to_pos, from_buff->buffer[i],
		   from_buff->nsamples * sizeof(short));
	}
    }

    to_buff->nsamples = nsamples;

    return from_buff->nsamples;
}

Bool is_audio_buffer_empty(AudioBuffer* buff)
{
    return buff == NULL ||
	buff->buffer == NULL ||
	    buff->nsamples == 0;
}

Bool merge_audio_buffer(AudioBuffer* from_buff, AudioBuffer* to_buff,
			long to_pos)
{
    int nch, sr, i;
    long nsamples, j;

    if(is_audio_buffer_empty(from_buff))
	return True;		/* from buffer is empty */

    FIXREGION(to_pos, 0, to_buff->nsamples);
    if(is_audio_buffer_empty(to_buff) || to_buff->nsamples == to_pos)
	if(insert_audio_buffer(from_buff, to_buff, to_pos) >= 0)
	    return True;
	else
	    return False;

    nch = to_buff->nchannels;
    sr  = to_buff->sample_rate;

    if(from_buff->nchannels != nch || from_buff->sample_rate != sr)
    {
	AudioBuffer* temp;

	temp = get_audio_buffer(AUDIO_TEMP_BUFF_NAME);
	if(convert_audio_format(from_buff, temp,
				0, from_buff->nsamples,
				nch, sr) == False)
	    return False;
	from_buff = temp;
    }

    nsamples = from_buff->nsamples;
    audio_buffer_extend_buffer(to_buff, to_pos + nsamples);
    for(i = 0; i < nch; i++)
    {
	short *to, *from;
	long x;

	to = to_buff->buffer[i] + to_pos;
	from = from_buff->buffer[i];
	for(j = 0; j < nsamples; j++)
	{
	    x = to[j] + from[j];
	    if(x < -32768)
		to[j] = -32768;
	    else if(x > 32767)
		to[j] = 32767;
	    else
		to[j] = (short)x;
	}
    }
    return True;
}

Bool overwrite_audio_buffer(AudioBuffer* from_buff, AudioBuffer* to_buff,
			    long to_pos)
{
    int nch, sr, i;
    long nsamples, j;

    if(is_audio_buffer_empty(from_buff))
	return True;		/* from buffer is empty */

    FIXREGION(to_pos, 0, to_buff->nsamples);
    if(is_audio_buffer_empty(to_buff) || to_buff->nsamples == to_pos)
	if(insert_audio_buffer(from_buff, to_buff, to_pos) >= 0)
	    return True;
	else
	    return False;

    nch = to_buff->nchannels;
    sr  = to_buff->sample_rate;

    if(from_buff->nchannels != nch || from_buff->sample_rate != sr)
    {
	AudioBuffer* temp;

	temp = get_audio_buffer(AUDIO_TEMP_BUFF_NAME);
	if(convert_audio_format(from_buff, temp,
				0, from_buff->nsamples,
				nch, sr) == False)
	    return False;
	from_buff = temp;
    }

    audio_buffer_extend_buffer(to_buff, to_pos + from_buff->nsamples);
    nsamples = from_buff->nsamples;
    for(i = 0; i < nch; i++)
    {
	short *to, *from;

	to = to_buff->buffer[i] + to_pos;
	from = from_buff->buffer[i];
	for(j = 0; j < nsamples; j++)
	    to[j] = from[j];
    }
    return True;
}
