#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fft.h"

#ifdef DEBUG
#ifdef sun
int fprintf(FILE *, const char *, ...);
#endif
#endif


#define swap(x, j, k)	\
{			\
    double tmp;		\
    tmp = x[j];		\
    x[j] = x[k];	\
    x[k] = tmp;		\
}

#define swap2(x, j, k)	\
{			\
    double tmp;		\
    tmp = x[j];		\
    x[j] = -x[k];	\
    x[k] = tmp;		\
}

static void make_table(double* trig, int* b, int n)
{
    unsigned long i, j, k, bitmask;

    /* check n */
    for(i = n; !(i & 1); i >>= 1)
	;
    if(i != 1)
    {
	fprintf(stderr, "Invalid fft data size: %d\n", n);
	exit(1);
    }

    /* make bitrev table */
    for(i = 0; i < n; i++)
	b[i] = 0;
    bitmask = n / 2;
    for(i = 1; i < n; i <<= 1, bitmask >>= 1)
	for(j = 0; j < n; j += i * 2)
	    for(k = j + i; k < j + i * 2; k++)
		b[k] = (int)((unsigned long)b[k] | bitmask);

    /* make trig table */
    for(i = 0; i < n; i++)
    {
	j = i * 2;
	trig[j  ] = cos((2 * M_PI) * i / (double)n);
	trig[j+1] = sin((2 * M_PI) * i / (double)n);
    }

    for(i = 0; i < n; i++)
	if(i < b[i])
	{
	    j = b[i] * 2;
	    swap(trig, i*2, j);
	    swap(trig, i*2+1, j+1);
	}

}

void realfft(double* x, int n_arg)
{
    int n = n_arg;
    static double* trig_table = NULL;
    static int* bitrev_table;
    int n1;

    if(n == 0)
    {
	if(trig_table == NULL)
	    return;
	free(trig_table);
	free(bitrev_table);
	trig_table = NULL;
	return;
    }

    if(trig_table == NULL)
    {
	trig_table = (double *)malloc((n * 2) * sizeof(double));
	bitrev_table = (int *)malloc(n * sizeof(int));

	if(!(trig_table && bitrev_table))
	{
	    fprintf(stderr, "fft: Can't allocate memroy.\n");
	    exit(1);
	}

	make_table(trig_table, bitrev_table, n);

	if(x == NULL)		/* initialize only */
	    return;
    }
    /* end of initialize tables */


    /* first step: butterfly w^0 */
    /* n1 = n/2, n/4, n/8, ... 1 */
    for(n1 = n >> 1; n1 > 0; n1 >>= 1)
    {
	int i;

	for(i = 0; i < n1; i++)
	{
	    double x1, x2;

	    x[i]    = (x1 = x[i]) + (x2 = x[i+n1]);
	    x[i+n1] = x1 - x2;
	}
    }

    /* main loop */
    /* n1 = n/8, n/16, n/32, ... 1 */
    for(n1 = n >> 3; n1 > 0; n1 >>= 1)
    {
	int i;
	int wi0;

	wi0 = 8;
	for(i = n1 << 2; i < n; i <<= 1, wi0 <<= 1)
	{
	    double* imp;	/* pointer to start of imaginary part */
	    double* rep;	/* pointer to start of real part  */
	    double* imp2;	/* pointer to start of imaginary part */
	    double* rep2;	/* pointer to start of real part  */
	    int wi, j0;
	    int in;

	    in   = i >> 1;
	    wi   = wi0;
	    rep  = x + i;
	    imp  = rep + in;
	    rep2 = rep + n1;
	    imp2 = imp + n1;

	    for(j0 = 0; j0 < in; j0 += (n1 << 1), wi += 4)
	    {
		int j, jn;
		double c, s;

		c  = trig_table[wi];
		s  = trig_table[wi+1];
		jn = j0 + n1;

		/* butterfly */
		for(j = j0; j < jn; j++)
		{
		    double xi1, xr1, xi2, xr2, ti, tr;

		    rep[j]  = (xr1 = rep[j]) + (tr = c * (xr2 = rep2[j])
					     - s * (xi2 = imp2[j]));
		    imp[j]  = (xi1 = imp[j]) + (ti = s * xr2 + c * xi2);
		    rep2[j] = xr1 - tr;
		    imp2[j] = xi1 - ti;
		}
	    }
	}
    }

    /* move data */
    {
	int ii, i, j, i2, i4;

	for(i = 4; i < n; i = ii)
	{
	    i2 = i >> 1;
	    i4 = i2 >> 1;
	    ii = i << 1;
	    for(j = 0; j < i4; j++)
		swap(x, j + i + i2, ii - j - 1);
	    for(j = 1; j < i2; j += 2)
		swap2(x, j + i, ii - j - 1);
	}

	for(i = 0; i < n; i++)
	    if(i < bitrev_table[i])
		swap(x, i, bitrev_table[i]);
	for(i = n/2+1; i < n; i++)
	    x[i] = -x[i];
    }
}

static void make_table2(double* trig, double* itrig, int* b, int n)
{
    unsigned long i, j, k, bitmask;

    /* check n */
    for(i = n; !(i & 1); i >>= 1)
	;
    if(i != 1)
    {
	fprintf(stderr, "Invalid fft data size: %d\n", n);
	exit(1);
    }

    /* make bitrev table */
    for(i = 0; i < n; i++)
	b[i] = 0;
    bitmask = n / 2;
    for(i = 1; i < n; i <<= 1, bitmask >>= 1)
	for(j = 0; j < n; j += i * 2)
	    for(k = j + i; k < j + i * 2; k++)
		b[k] = (int)((unsigned long)b[k] | bitmask);

    /* make trig table */
    for(i = 0; i < n; i++)
    {
	double r = (2 * M_PI) * i / (double)n;
	j = i * 2;
	itrig[j  ] = trig[j  ] = cos(r);
	itrig[j+1] = -(trig[j+1] = sin(r));
    }

    for(i = 0; i < n; i++)
    {
	if(i < b[i])
	{
	    double a, bb, c, d;
	    unsigned long i2, j2;

	    i2 = i * 2;
	    j2 = b[i] * 2;

	    a = trig[i2  ];
	    bb = trig[i2+1];
	    c = itrig[i2  ];
	    d = itrig[i2+1];

	    trig[i2  ]  = trig[j2  ];
	    trig[i2+1]  = trig[j2+1];
	    itrig[i2  ] = itrig[j2  ];
	    itrig[i2+1] = itrig[j2+1];

	    trig[j2  ] = a;
	    trig[j2+1] = bb;
	    itrig[j2  ] = c;
	    itrig[j2+1] = d;
	}
    }
}

void fft(double* x, double* y, int n)
{
    static double* trig_table = NULL;
    static double* itrig_table;
    static double* trig;
    static int* bitrev_table;
    static int init_n;
    int i;

    if(n == 0)
    {
	if(trig_table == NULL)
	    return;
	free(trig_table);
	free(itrig_table);
	free(bitrev_table);
	trig_table = NULL;
	return;
    }

    if(trig_table == NULL)
    {
	if(n < 0)
	    init_n = -n;
	else
	    init_n = n;
	trig_table = (double *)malloc((init_n * 2) * sizeof(double));
	itrig_table = (double *)malloc((init_n * 2) * sizeof(double));
	bitrev_table = (int *)malloc(init_n * sizeof(int));

	if(!(trig_table && itrig_table && bitrev_table))
	{
	    fprintf(stderr, "fft: Can't allocate memroy.\n");
	    exit(1);
	}

	make_table2(trig_table, itrig_table, bitrev_table, init_n);

	if(x == NULL || y == NULL)	/* initialize only */
	    return;
    }

    if(n > 0)
	trig = itrig_table;
    else
    {
	trig = trig_table;
	n = -n;
    }

    if(n != init_n)
    {
	fprintf(stderr, "fft: Invalid number of data: %d\n", n);
	exit(1);
    }

    /* main fft proc */
    {
	int j, k;
	int i2, bi;
	int m, ji;
	double c, s;
	double t1x, t1y, t2x, t2y, tx, ty;

	i2 = n / 2;
	for(i = n; i > 1; i = i2, i2 >>= 1)
	{
	    bi = 0;
	    for(j = 0; j < n; j += i, bi += 2)
	    {
		c = trig[bi * 2];
		s = trig[bi * 2 + 1];

		ji = j + i;
		m = j;
		for(k = j + i2; k < ji; k++, m++)
		{
		    t1x = x[m];
		    t1y = y[m];
		    tx = x[k];
		    ty = y[k];
		    t2x = c * tx - s * ty;
		    t2y = s * tx + c * ty;
		    x[m] = t1x + t2x;
		    y[m] = t1y + t2y;
		    x[k] = t1x - t2x;
		    y[k] = t1y - t2y;
		}
	    }
	}
    }

    /* move data */
    for(i = 0; i < n; i++)
    {
	if(i < bitrev_table[i])
	{
	    double tmp;

	    tmp = x[i];
	    x[i] = x[bitrev_table[i]];
	    x[bitrev_table[i]] = tmp;
	    tmp = y[i];
	    y[i] = y[bitrev_table[i]];
	    y[bitrev_table[i]] = tmp;
	}
    }

    if(trig == itrig_table)
    {
	double r = 1.0 / n;

	for(i = 0; i < n; i++)
	{
	    x[i] *= r;
	    y[i] *= r;
	}
    }
}
