#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/param.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "usffd.h"
#include "ieee.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN      1024
#endif

/* Sun Audio File Encoding Tags */
#define SUN_AUDIO_MAGIC			0x2e736e64
#define AUDIO_FILE_ENCODING_MULAW_8     1       /* 8-bit ISDN u-law */
#define AUDIO_FILE_ENCODING_ALAW_8     27       /* 8-bit ISDN A-law */
#define AUDIO_FILE_ENCODING_LINEAR_8    2       /* 8-bit linear PCM */
#define AUDIO_FILE_ENCODING_LINEAR_16   3       /* 16-bit linear PCM */

/* Not supported */
#define AUDIO_FILE_ENCODING_LINEAR_24   4      /* 24-bit linear PCM */
#define AUDIO_FILE_ENCODING_LINEAR_32   5      /* 32-bit linear PCM */
#define AUDIO_FILE_ENCODING_FLOAT       6      /* 32-bit IEEE floating point */
#define AUDIO_FILE_ENCODING_DOUBLE      7      /* 64-bit IEEE floating point */
#define AUDIO_FILE_ENCODING_INDIRECT    8      /* fragmented sampled data */
#define AUDIO_FILE_ENCODING_NESTED      9      /* ? */
#define AUDIO_FILE_ENCODING_DSP_CORE   10      /* DSP program */
#define AUDIO_FILE_ENCODING_DSP_DATA_8 11      /* 8-bit fixed-point samples */
#define AUDIO_FILE_ENCODING_DSP_DATA_16 12     /* 16-bit fixed-point samples */
#define AUDIO_FILE_ENCODING_DSP_DATA_24 13     /* 24-bit fixed-point samples */
#define AUDIO_FILE_ENCODING_DSP_DATA_32 14     /* 32-bit fixed-point samples */
#define AUDIO_FILE_ENCODING_DISPLAY     16     /* non-audio display data */
#define AUDIO_FILE_ENCODING_MULAW_SQUELCH 17   /* ? */
#define AUDIO_FILE_ENCODING_EMPHASIZED 18      /* 16-bit linear with
						  emphasis */
#define AUDIO_FILE_ENCODING_COMPRESSED 19      /* 16-bit linear with
						  compression */
#define AUDIO_FILE_ENCODING_COMPRESSED_EMPHASIZED 20 /* A combination of the
						        two above */
#define AUDIO_FILE_ENCODING_DSP_COMMANDS 21    /*  Music Kit DSP commands */
#define AUDIO_FILE_ENCODING_DSP_COMMANDS_SAMPLES 22 /* ? */

#define AUDIO_FILE_ENCODING_ADPCM_G721  23     /* 4-bit CCITT g.721 ADPCM */
#define AUDIO_FILE_ENCODING_ADPCM_G722  24     /* CCITT g.722 ADPCM */
#define AUDIO_FILE_ENCODING_ADPCM_G723_3 25    /* CCITT g.723 3-bit ADPCM */
#define AUDIO_FILE_ENCODING_ADPCM_G723_5 26    /* CCITT g.723 5-bit ADPCM */




/* Windows WAVE File Encoding Tags */
#define WAVE_FORMAT_PCM       0x0001
#define WAVE_FORMAT_ADPCM     0x0002 /* Not suppoted */
#define WAVE_FORMAT_ALAW      0x0006
#define WAVE_FORMAT_MULAW     0x0007
#define WAVE_FORMAT_OKI_ADPCM 0x0010 /* Not suppoted */
#define WAVE_FORMAT_DIGISTD   0x0015 /* Not suppoted */
#define WAVE_FORMAT_DIGIFIX   0x0016 /* Not suppoted */
#define IBM_FORMAT_MULAW      0x0101 /* Not suppoted */
#define IBM_FORMAT_ALAW       0x0102 /* Not suppoted */
#define IBM_FORMAT_ADPCM      0x0103 /* Not suppoted */

#define DEBUG

#ifdef DEBUG
#ifdef sun
int fprintf(FILE *, const char *, ...);
#endif
#endif

int usffd_errno = USFFD_NO_ERROR;
long usffd_default_sample_rate = 8000;

const int usffd_encoding_sample_size[USFFD_NENCODINGS] = {
    0, /* USFFD_UNDEFINED */
    1, /* USFFD_SIGBYTE */
    1, /* USFFD_UNSIGBYTE */
    1, /* USFFD_G711_ULAW */
    1, /* USFFD_G711_ALAW */
    2, /* USFFD_SIGWORDB */
    2, /* USFFD_UNSIGWORDB */
    2, /* USFFD_SIGWORDL */
    2, /* USFFD_UNSIGWORDL */
};


static void usffd_default_encoding_auto_recognizer(USFFD* usffd)
{
    fprintf(stderr, "WARNING: Unrecognized sound file encording `%s'\n",
	    usffd->filename);
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_G711_ULAW;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 1;
    usffd->hdr_type = USFFD_HDR_NONE;
}

void (* usffd_encoding_auto_recognizer)(USFFD* usffd) =
    usffd_default_encoding_auto_recognizer;

static Bool usffd_set_ub_header(USFFD* usffd);
static Bool usffd_set_ub2_header(USFFD* usffd);
static Bool usffd_set_sb_header(USFFD* usffd);
static Bool usffd_set_sb2_header(USFFD* usffd);
static Bool usffd_set_uw_header(USFFD* usffd);
static Bool usffd_set_uw2_header(USFFD* usffd);
static Bool usffd_set_sw_header(USFFD* usffd);
static Bool usffd_set_sw2_header(USFFD* usffd);
static Bool usffd_set_ul_header(USFFD* usffd);
static Bool usffd_set_ul2_header(USFFD* usffd);
static Bool usffd_set_al_header(USFFD* usffd);
static Bool usffd_set_al2_header(USFFD* usffd);
static Bool usffd_set_au_header(USFFD* usffd);
static Bool usffd_set_wav_header(USFFD* usffd);
static Bool usffd_set_aiff_header(USFFD* usffd);

static struct usffd_setting_header_handler
{
    char* file_extention;
    Bool need_file_header;
    Bool (* usffd_read_header)(USFFD* usffd);
} usffd_setting_header_handler[] =
{
{"sb",	False,	usffd_set_sb_header},	/* signed byte */
{"sb2",	False,	usffd_set_sb2_header},	/* signed byte */
{"ub",	False,	usffd_set_ub_header},	/* unsigned byte */
{"ub2",	False,	usffd_set_ub2_header},	/* unsigned byte */
{"uw",	False,	usffd_set_uw_header},	/* unsigned word */
{"uw2",	False,	usffd_set_uw2_header},	/* unsigned word */
{"sw",	False,	usffd_set_sw_header},	/* signed word */
{"sw2",	False,	usffd_set_sw2_header},	/* signed word */
{"ul",	False,	usffd_set_ul_header},	/* ulaw */
{"ul2",	False,	usffd_set_ul2_header},	/* ulaw */
{"al",	False,	usffd_set_al_header},	/* alaw */
{"al2",	False,	usffd_set_al2_header},	/* alaw */
{"au",	True,	usffd_set_au_header},	/* SunAudio */
{"wav",	True,	usffd_set_wav_header},	/* Windows WAVE */
{"WAV",	True,	usffd_set_wav_header},	/* Windows WAVE */
{"aiff",True,   usffd_set_aiff_header}, /* SGI AIFF */
{"aif", True,   usffd_set_aiff_header}, /* SGI AIFF */
{NULL,  False,  NULL}/* terminate */
};

void usffd_print(const USFFD* usffd)
{
    printf("USFFD: fd = %d\n", usffd->fd);
    printf("USFFD: filename = %s\n", usffd->filename);
    printf("USFFD: file_size = %ld\n", usffd->file_size);
    printf("USFFD: hdr_size = %ld\n", usffd->hdr_size);
    printf("USFFD: data_size = %ld\n", usffd->data_size);
    printf("USFFD: data_encoding = %ld\n", usffd->data_encoding);
    printf("USFFD: sample_rate = %ld\n", usffd->sample_rate);
    printf("USFFD: nchannels = %ld\n", usffd->nchannels);
    printf("USFFD: default file extension = \"%s\"\n",
	   usffd_hdr_extension(usffd));
}

Bool usffd_open(const char* filename, USFFD* usffd)
{
    struct stat buf;

    usffd_errno = USFFD_NO_ERROR;

    usffd->fd = -1;
    usffd->filename = NULL;
    usffd->file_size = 0;
    usffd->hdr_size = 0;
    usffd->data_size = 0;
    usffd->data_encoding = USFFD_UNDEFINED;
    usffd->sample_rate = USFFD_UNDEFINED;
    usffd->nchannels = 1;
    usffd->hdr_type = USFFD_HDR_NONE;

    if((usffd->filename = strdup(filename)) == NULL)
    {
	usffd_errno = USFFD_SYS_ERROR;
	return False;
    }

    if((usffd->fd = open(filename, O_RDONLY)) < 0)
    {
	usffd_errno = USFFD_SYS_ERROR;
	free(usffd->filename);
	usffd->filename = NULL;
	return False;
    }

    if(fstat(usffd->fd, &buf) < 0)
    {
	usffd_errno = USFFD_SYS_ERROR;
	close(usffd->fd);
	free(usffd->filename);
	usffd->fd = -1;
	usffd->filename = NULL;
	return False;
    }

    if(!S_ISREG(buf.st_mode))
    {
	usffd_errno = USFFD_NOT_REGULAR_FILE;
	close(usffd->fd);
	free(usffd->filename);
	usffd->fd = -1;
	usffd->filename = NULL;
	return False;
    }

    usffd->file_size = buf.st_size;

    return True;
}

void usffd_make_header(USFFD* usffd,
		       long encoding, long nchannels, long sample_rate)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = encoding;
    usffd->sample_rate = sample_rate;
    usffd->nchannels = nchannels;
    usffd->hdr_type = USFFD_HDR_NONE;
}

void usffd_close(USFFD* usffd)
{
    if(usffd->fd >= 0)
    {
	close(usffd->fd);
	usffd->fd = -1;
    }

    if(usffd->filename != NULL)
    {
	free(usffd->filename);
	usffd->filename = NULL;
    }

    usffd->file_size = 0;
    usffd->hdr_size = 0;
    usffd->data_size = 0;
    usffd->data_encoding = USFFD_UNDEFINED;
    usffd->sample_rate = USFFD_UNDEFINED;
    usffd->nchannels = USFFD_UNDEFINED;
    usffd->hdr_type = USFFD_HDR_NONE;
}

Bool usffd_set_header(USFFD* usffd)
{
    int i;
    char* p;

    usffd_errno = USFFD_NO_ERROR;

    p = usffd->filename + strlen(usffd->filename);
    while(usffd->filename < p && *p != '.')
	p--;
    if(usffd->filename < p && *p == '.')
    {
	p++;
	for(i = 0; usffd_setting_header_handler[i].file_extention != NULL; i++)
	    if(!strcmp(usffd_setting_header_handler[i].file_extention, p))
		return usffd_setting_header_handler[i].
		    usffd_read_header(usffd);
    }

    return False;
}

void usffd_encoding_recognize(USFFD* usffd)
{
    int i;

    /* ޤǽˡեإåĴ٤롥*/
    for(i = 0; usffd_setting_header_handler[i].file_extention != NULL; i++)
    {
	lseek(usffd->fd, 0, SEEK_SET);
	if(usffd_setting_header_handler[i].need_file_header &&
	   usffd_setting_header_handler[i].usffd_read_header(usffd))
	{
	    usffd_errno = USFFD_NO_ERROR;
	    return;
	}
    }
    usffd_errno = USFFD_NO_ERROR;
    lseek(usffd->fd, 0, SEEK_SET);

    /* ٤Ƽԡưǧư */
    usffd_encoding_auto_recognizer(usffd);
    lseek(usffd->fd, usffd->hdr_size, SEEK_SET);
}

static U32 read_ulong_bigend(int fd)
{
#ifdef BIGENDIAN
    U32 val;

    if(read(fd, &val, 4) != 4)
	return 0;

    return val;
#else
    unsigned char c[4];

    if(read(fd, c, 4) != 4)
	return 0;

    return  ((U32)c[3])		|
	    (((U32)c[2]) << 8)	|
	    (((U32)c[1]) << 16)	|
	    (((U32)c[0]) << 24);
#endif
}

static void write_ulong_bigend(int fd, U32 value)
{
#ifndef BIGENDIAN
    value = ((value << 24) |
	     ((value & 0x0000ff00) << 8) |
	     ((value & 0x00ff0000) >> 8) |
	     (value >> 24));
#endif
    write(fd, &value, 4);
}

static U32 read_ulong_littleend(int fd)
{
#ifdef BIGENDIAN
    unsigned char c[4];

    if(read(fd, c, 4) != 4)
	return 0;

    return  ((U32)c[0])		|
	    (((U32)c[1]) << 8)	|
	    (((U32)c[2]) << 16)	|
	    (((U32)c[3]) << 24);
#else
    U32 val;

    if(read(fd, &val, 4) != 4)
      return 0;
    return val;
#endif
}

static void write_ulong_littleend(int fd, U32 value)
{
#ifdef BIGENDIAN
    value = ((value << 24) |
	     ((value & 0x0000ff00) << 8) |
	     ((value & 0x00ff0000) >> 8) |
	     (value >> 24));
#endif
    write(fd, &value, 4);
}

static U16 read_ushort_littleend(int fd)
{
#ifdef BIGENDIAN
    unsigned char c[2];

    if(read(fd, c, 2) != 2)
	return 0;

    return ((U16)c[0]) | (((U16)c[1]) << 8);
#else
    U16 val;

    if(read(fd, &val, 2) != 2)
      return 0;

    return val;
#endif /* BIGENDIAN */
}

static U16 read_ushort_bigend(int fd)
{
#ifdef BIGENDIAN
    U16 val;

    if(read(fd, &val, 2) != 2)
      return 0;

    return val;
#else
    unsigned char c[2];

    if(read(fd, c, 2) != 2)
	return 0;

    return ((U16)c[1]) | (((U16)c[0]) << 8);
#endif /* BIGENDIAN */
}

static void write_ushort_littleend(int fd, U16 value)
{
#ifdef BIGENDIAN
    value = ((value << 8) | (value >> 8));
#endif
    write(fd, &value, 2);
}

static Bool usffd_set_ub_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_UNSIGBYTE;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 1;

    return True;
}

static Bool usffd_set_ub2_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_UNSIGBYTE;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 2;

    return True;
}

static Bool usffd_set_sb_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_SIGBYTE;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 1;

    return True;
}

static Bool usffd_set_sb2_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_SIGBYTE;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 2;

    return True;
}

static Bool usffd_set_uw_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size - usffd->file_size % 2;
#ifdef BIGENDIAN
    usffd->data_encoding = USFFD_UNSIGWORDB;
#else
    usffd->data_encoding = USFFD_UNSIGWORDL;
#endif
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 1;

    return True;
}

static Bool usffd_set_uw2_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size - usffd->file_size % 2;
#ifdef BIGENDIAN
    usffd->data_encoding = USFFD_UNSIGWORDB;
#else
    usffd->data_encoding = USFFD_UNSIGWORDL;
#endif
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 2;

    return True;
}

static Bool usffd_set_sw_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size - usffd->file_size % 2;
#ifdef BIGENDIAN
    usffd->data_encoding = USFFD_SIGWORDB;
#else
    usffd->data_encoding = USFFD_SIGWORDL;
#endif
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 1;

    return True;
}

static Bool usffd_set_sw2_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size - usffd->file_size % 2;
#ifdef BIGENDIAN
    usffd->data_encoding = USFFD_SIGWORDB;
#else
    usffd->data_encoding = USFFD_SIGWORDL;
#endif
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 2;

    return True;
}

static Bool usffd_set_ul_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_G711_ULAW;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 1;

    return True;
}

static Bool usffd_set_ul2_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_G711_ULAW;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 2;

    return True;
}

static Bool usffd_set_al_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_G711_ULAW;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 1;

    return True;
}

static Bool usffd_set_al2_header(USFFD* usffd)
{
    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    usffd->data_encoding = USFFD_G711_ULAW;
    usffd->sample_rate = usffd_default_sample_rate;
    usffd->nchannels = 2;

    return True;
}

static Bool usffd_set_au_header(USFFD* usffd)
{
    struct au_hdr
    {
	U32 magic;
	U32 hdr_size;
	U32 data_size;
	U32 data_encoding;
	U32 sample_rate;
	U32 nchannels;
    } hdr;

    hdr.magic         = read_ulong_bigend(usffd->fd);
    hdr.hdr_size      = read_ulong_bigend(usffd->fd);
    hdr.data_size     = read_ulong_bigend(usffd->fd);
    hdr.data_encoding = read_ulong_bigend(usffd->fd);
    hdr.sample_rate   = read_ulong_bigend(usffd->fd);
    hdr.nchannels     = read_ulong_bigend(usffd->fd);

#ifdef DEBUG
    printf("hdr.magic: 0x%08x\n", hdr.magic);
    printf("hdr.hdr_size: %d\n", hdr.hdr_size);
    printf("hdr.data_size: %d\n", hdr.data_size);
    printf("hdr.data_encoding: %d\n", hdr.data_encoding);
    printf("hdr.sample_rate: %d\n", hdr.sample_rate);
    printf("hdr.nchannels: %d\n", hdr.nchannels);
#endif

    if(hdr.magic != SUN_AUDIO_MAGIC)
    {
	usffd_errno = USFFD_INVALID_FILE_HEADER;
	lseek(usffd->fd, 0, SEEK_SET);
	return False;
    }

    usffd->hdr_size = hdr.hdr_size;
    usffd->data_size = hdr.data_size;
    if(usffd->data_size == 0xffffffff)
	usffd->data_size = usffd->file_size - hdr.hdr_size;

    switch(hdr.data_encoding)
    {
      case AUDIO_FILE_ENCODING_MULAW_8:
	usffd->data_encoding = USFFD_G711_ULAW;
	break;

      case AUDIO_FILE_ENCODING_ALAW_8:
	usffd->data_encoding = USFFD_G711_ALAW;
	break;

      case AUDIO_FILE_ENCODING_LINEAR_8:
	usffd->data_encoding = USFFD_SIGBYTE;
	break;

      case AUDIO_FILE_ENCODING_LINEAR_16:
	usffd->data_encoding = USFFD_SIGWORDB;
	break;

      default:
	usffd->data_encoding = USFFD_G711_ULAW;
	usffd_errno = USFFD_UNKNOWN_FILE_ENCODING;
	break;
    }

    usffd->sample_rate = hdr.sample_rate;
    usffd->nchannels = hdr.nchannels;
    usffd->hdr_type = USFFD_HDR_AU;

    lseek(usffd->fd, usffd->hdr_size, SEEK_SET);

    return True;
}

static Bool usffd_set_wav_header(USFFD* usffd)
{
    char buff[8];
    U32 bits_per_sample, size, chunk_size, len;
    U16 format;

    read(usffd->fd, buff, 4);	/* expect "RIFF" */
    if(strncmp(buff, "RIFF", 4) != 0)
    {
        lseek(usffd->fd, 0, SEEK_SET);
	usffd_errno = USFFD_INVALID_FILE_HEADER;
        return False;
    }

    size = read_ulong_littleend(usffd->fd);
    read(usffd->fd, buff, 4);	/* expect "WAVE" */
    if(strncmp(buff, "WAVE", 4) != 0)
    {
        lseek(usffd->fd, 0, SEEK_SET);
	usffd_errno = USFFD_INVALID_FILE_HEADER;
        return False;
    }

    /* Skip to the next "fmt " or end of file */
    while(1)
    {
	if(read(usffd->fd, buff, 4) != 4)
	{
	    lseek(usffd->fd, 0, SEEK_SET);
	    usffd_errno = USFFD_UNKNOWN_FILE_ENCODING;
	    return False;
	}
	if(!strncmp("fmt ", buff, 4))
	    break;
	len = read_ulong_littleend(usffd->fd);
	while(len > 0 && read(usffd->fd, buff, 1) > 0)
	    len--;
    }

    chunk_size = read_ulong_littleend(usffd->fd);
    format = read_ushort_littleend(usffd->fd);
    usffd->nchannels = read_ushort_littleend(usffd->fd);
    if(usffd->nchannels == 0)
    {
#ifdef DEBUG
	puts("usffd->nchannels == 0");
#endif
	usffd_errno = USFFD_INVALID_FILE_HEADER;
        return False;
    }
    usffd->sample_rate = read_ulong_littleend(usffd->fd);

    /* avg_bytes_per_sec */
    read_ulong_littleend(usffd->fd);

    /* block_align */
    read_ushort_littleend(usffd->fd);

    bits_per_sample = read_ushort_littleend(usffd->fd);
    if(bits_per_sample != 8 && bits_per_sample != 16)
    {
        lseek(usffd->fd, 0, SEEK_SET);
	usffd_errno = USFFD_UNKNOWN_FILE_ENCODING;
        return False;
    }

    chunk_size -= 16;
    while(chunk_size > 0 && read(usffd->fd, buff, 1) > 0)
	chunk_size--;

    /* Now look for the wave data chunk */
    for (;;)
    {
	if(read(usffd->fd, buff, 4) != 4)
	{
	    lseek(usffd->fd, 0, SEEK_SET);
	    usffd_errno = USFFD_UNKNOWN_FILE_ENCODING;
	    return False;
	}

	len = read_ulong_littleend(usffd->fd);
	if(strncmp("data", buff, 4) == 0)
	    break;	/* Found the data chunk */
	while(len > 0 && read(usffd->fd, buff, 1) > 0)
	    len--;
    }

    switch(format)
    {
      case WAVE_FORMAT_PCM:
	if(bits_per_sample == 8)
	    usffd->data_encoding = USFFD_UNSIGBYTE;
	else
	    usffd->data_encoding = USFFD_SIGWORDL;
	break;

      case WAVE_FORMAT_ALAW:
	if(bits_per_sample != 8)
	{
	    lseek(usffd->fd, 0, SEEK_SET);
	    usffd_errno = USFFD_INVALID_FILE_HEADER;
	    return False;
	}
	usffd->data_encoding = USFFD_G711_ALAW;
	break;

      case WAVE_FORMAT_MULAW:
	if(bits_per_sample != 8)
	{
	    lseek(usffd->fd, 0, SEEK_SET);
	    usffd_errno = USFFD_INVALID_FILE_HEADER;
	    return False;
	}
	usffd->data_encoding = USFFD_G711_ULAW;
	break;

      default:
	usffd->data_encoding = USFFD_G711_ULAW;
	usffd_errno = USFFD_UNKNOWN_FILE_ENCODING;
#ifdef DEBUG
	printf("format = %x\n", format);
#endif
	break;
    }

    usffd->hdr_size = lseek(usffd->fd, 0, SEEK_CUR);
    usffd->data_size = len;
    usffd->hdr_type = USFFD_HDR_WAV;

    return True;
}

static Bool usffd_set_aiff_header(USFFD* usffd)
{
    char buff[10];
    unsigned long size, chunksize, frames, offset;
    int channels;
    int bits;
    double rate;

    read(usffd->fd, buff, 4);	/* expect "FORM" */
    if(strncmp(buff, "FORM", 4) != 0)
    {
        lseek(usffd->fd, 0, SEEK_SET);
	usffd_errno = USFFD_INVALID_FILE_HEADER;
	return False;
    }
    size = read_ulong_bigend(usffd->fd);
    read(usffd->fd, buff, 4);	/* expect "AIFFCOMM" */
    if(strncmp(buff, "AIFF", 4) != 0)
    {
        lseek(usffd->fd, 0, SEEK_SET);
	usffd_errno = USFFD_INVALID_FILE_HEADER;
	return False;
    }

    while(1)
    {
	if(read(usffd->fd, buff, 4) != 4)
	{
	    lseek(usffd->fd, 0, SEEK_SET);
	    usffd_errno = USFFD_INVALID_FILE_HEADER;
	    return False;
	}

	if(strncmp(buff, "COMM", 4) == 0)
	{
	    chunksize = read_ulong_bigend(usffd->fd);
#if 0
	    if(chunksize != 18) /* or 38? */
	    {
		printf("## ERR chunksize=%d\n", chunksize);
		lseek(usffd->fd, 0, SEEK_SET);
		usffd_errno = USFFD_INVALID_FILE_HEADER;
		return False;
	    }
#endif
	    channels = read_ushort_bigend(usffd->fd);
	    frames = read_ulong_bigend(usffd->fd);
	    bits = read_ushort_bigend(usffd->fd);
	    if(read(usffd->fd, buff, 10) != 10)
	    {
		lseek(usffd->fd, 0, SEEK_SET);
		usffd_errno = USFFD_INVALID_FILE_HEADER;
		return False;
	    }
	    rate = ConvertFromIeeeExtended((unsigned char *)buff);
	}
	else if(strncmp(buff, "SSND", 4) == 0)
	{
	    unsigned long blocksize;

	    chunksize = read_ulong_bigend(usffd->fd);
	    offset    = read_ulong_bigend(usffd->fd);
	    blocksize = read_ulong_bigend(usffd->fd);
	    if(blocksize != 0)
	    {
		lseek(usffd->fd, 0, SEEK_SET);
		usffd_errno = USFFD_INVALID_FILE_HEADER;
		return False;
	    }

	    /* skip offset bytes */
	    lseek(usffd->fd, offset, SEEK_CUR);
	    break;
	}
	else
	{
	    chunksize = read_ulong_bigend(usffd->fd);
	    lseek(usffd->fd, chunksize, SEEK_CUR);
	}
    }

    usffd->nchannels = channels;
    usffd->sample_rate = (long)rate;
    switch(bits)
    {
      case 8:
	usffd->data_encoding = USFFD_SIGBYTE;
	break;
      case 16:
	usffd->data_encoding = USFFD_SIGWORDB;
	break;
      default:
	lseek(usffd->fd, 0, SEEK_SET);
	usffd_errno = USFFD_INVALID_FILE_HEADER;
	return False;
    }

    usffd->data_size = chunksize - offset - 8;
    usffd->hdr_size  = lseek(usffd->fd, 0, SEEK_CUR);
    return True;
}

const char* usffd_encoding_jname(long encoding)
{
    switch(encoding)
    {
      case USFFD_SIGBYTE: return "դ1Х";
      case USFFD_UNSIGBYTE: return "ʤ1Х";
      case USFFD_G711_ULAW: return "ˡ§";
      case USFFD_G711_ALAW: return "ˡ§";

      case USFFD_SIGWORDB: return "դ2Х(big-endian)";
      case USFFD_UNSIGWORDB: return "ʤ2Х(big-endian)";
      case USFFD_SIGWORDL: return "դ2Х(little-endian)";
      case USFFD_UNSIGWORDL: return "ʤ2Х(little-endian)";
    }
    return "̤";
}

const char* usffd_encoding_name(long encoding)
{
    switch(encoding)
    {
      case USFFD_SIGBYTE: return "signed byte";
      case USFFD_UNSIGBYTE: return "unsigned byte";
      case USFFD_G711_ULAW: return "U-Law";
      case USFFD_G711_ALAW: return "A-Law";

      case USFFD_SIGWORDB: return "signed word(big-endian)";
      case USFFD_UNSIGWORDB: return "unsigned word(big-endian)";
      case USFFD_SIGWORDL: return "signed word(little-endian)";
      case USFFD_UNSIGWORDL: return "unsigned word(little-endian)";
    }
    return "undefined";
}

const char* usffd_strerror(int error_code)
{
  //extern char* sys_errlist[];
    static const char* usffd_error_text[] =
    {
	"",			/* USFFD_NO_ERROR */
	"",			/* USFFD_SYS_ERROR */
	"Not a regular file",	/* USFFD_NOT_REGULAR_FILE */
	"Invalid audio file header", /* USFFD_INVALID_FILE_HEADER */
	"Unrecognized file header", /* USFFD_UNRECOGNIZED_FILE_HEADER */
	"Unknown file encoding", /* USFFD_UNKNOWN_FILE_ENCODING */
    };

    if(error_code == USFFD_SYS_ERROR)
	return sys_errlist[errno];

    return usffd_error_text[error_code];
}

const char* usffd_hdr_extension(const USFFD* usffd)
{
    switch(usffd->hdr_type)
    {
      case USFFD_HDR_NONE:
	switch(usffd->data_encoding)
	{
	  case USFFD_SIGBYTE:
	    return usffd->nchannels == 2 ? ".sb2" : ".sb";
	  case USFFD_UNSIGBYTE:
	    return usffd->nchannels == 2 ? ".ub2" : ".ub";
	  case USFFD_G711_ULAW:
	    return usffd->nchannels == 2 ? ".ul2" : ".ul";
	  case USFFD_G711_ALAW:
	    return usffd->nchannels == 2 ? ".al2" : ".al";
	  case USFFD_SIGWORDB:
	    return usffd->nchannels == 2 ? ".sw2" : ".sw";
	  case USFFD_UNSIGWORDB:
	    return usffd->nchannels == 2 ? ".uw2" : ".uw";
	  case USFFD_SIGWORDL:
	    return usffd->nchannels == 2 ? ".sw2" : ".sw";
	  case USFFD_UNSIGWORDL:
	    return usffd->nchannels == 2 ? ".uw2" : ".uw";
	}
      case USFFD_HDR_AU:
	return ".au";
      case USFFD_HDR_WAV:
	return ".wav";
      case USFFD_HDR_AIFF:
	return ".aiff";
    }
    return "";
}

static U32 approximate_value(U32 value, const U32* value_list, int n)
{
    int i;
    U32 d1, d2;

    if(n == 1 || value <= value_list[0])
	return value_list[0];

    if(value >= value_list[n - 1])
	return value_list[n - 1];

    for(i = 0; i < n - 1; i++)
    {
	if(value_list[i] < value && value <= value_list[i + 1])
	    break;
    }

    d1 = value - value_list[i];
    d2 = value_list[i + 1] - value;

    if(d1 < d2)
	return value_list[i];
    return value_list[i + 1];
}

static U32 fix_sample_rate(U32 sample_rate)
{
    U32 sample_rate_list[] =
    {
	5512, 6615, 8000, 9600, 11025, 16000, 18900, 22050, 32000, 37800,
	44100, 48000
    };
    return approximate_value(sample_rate, sample_rate_list, 12);
}

static void usffd_write_au_header(int fd, const USFFD* usffd,
				  long* file_encoding)
{
    char* comment;
    int clen;
    struct au_hdr
    {
	U32 magic;
	U32 hdr_size;
	U32 data_size;
	U32 data_encoding;
	U32 sample_rate;
	U32 nchannels;
    } hdr;

    if((comment = usffd->filename) == NULL)
	comment = "";
    clen = strlen(comment);

    hdr.magic = SUN_AUDIO_MAGIC;
    hdr.hdr_size = 48 + clen;
    hdr.sample_rate = fix_sample_rate(usffd->sample_rate);
    hdr.nchannels = usffd->nchannels;
    hdr.data_size = usffd->data_size;

    switch(usffd->data_encoding)
    {
      case USFFD_SIGBYTE:
      case USFFD_UNSIGBYTE:
	*file_encoding = USFFD_SIGBYTE;
	hdr.data_encoding = AUDIO_FILE_ENCODING_LINEAR_8;
	break;

      case USFFD_G711_ULAW:
	*file_encoding = USFFD_G711_ULAW;
	hdr.data_encoding = AUDIO_FILE_ENCODING_MULAW_8;
	break;

      case USFFD_G711_ALAW:
	*file_encoding = USFFD_G711_ALAW;
	hdr.data_encoding = AUDIO_FILE_ENCODING_ALAW_8;
	break;

      case USFFD_SIGWORDB:
      case USFFD_UNSIGWORDB:
      case USFFD_SIGWORDL:
      case USFFD_UNSIGWORDL:
	*file_encoding = USFFD_SIGWORDB;
	hdr.data_encoding = AUDIO_FILE_ENCODING_LINEAR_16;
	break;
    }

    write_ulong_bigend(fd, hdr.magic);
    write_ulong_bigend(fd, hdr.hdr_size);
    write_ulong_bigend(fd, hdr.data_size);
    write_ulong_bigend(fd, hdr.data_encoding);
    write_ulong_bigend(fd, hdr.sample_rate);
    write_ulong_bigend(fd, hdr.nchannels);
    write(fd, comment, clen);
}

static void usffd_write_wav_header(int fd, const USFFD* usffd,
				   long* file_encoding)
{
    U16 format = 0, nchannels, bits_per_sample;
    U32 data_size, sample_rate, block_align, av_bps;

    data_size = usffd->data_size;
    nchannels = usffd->nchannels;
    bits_per_sample = 8 * usffd_encoding_sample_size[usffd->data_encoding];
    sample_rate = fix_sample_rate(usffd->sample_rate);
    block_align = nchannels * (bits_per_sample / 8);
    av_bps = block_align * sample_rate;

    switch(usffd->data_encoding)
    {
      case USFFD_SIGBYTE:
      case USFFD_UNSIGBYTE:
	*file_encoding = USFFD_UNSIGBYTE;
	format = WAVE_FORMAT_PCM;
	break;
      case USFFD_G711_ULAW:
	*file_encoding = USFFD_G711_ULAW;
	format = WAVE_FORMAT_MULAW;
	break;
      case USFFD_G711_ALAW:
	*file_encoding = USFFD_G711_ALAW;
	format = WAVE_FORMAT_ALAW;
	break;
      case USFFD_SIGWORDB:
      case USFFD_UNSIGWORDB:
      case USFFD_SIGWORDL:
      case USFFD_UNSIGWORDL:
	*file_encoding = USFFD_SIGWORDL;
	format = WAVE_FORMAT_PCM;
	break;
    }

    write(fd, "RIFF", 4);
    write_ulong_littleend(fd, data_size + 8+16+12);
    write(fd, "WAVEfmt ", 8);
    write_ulong_littleend(fd, 16);
    write_ushort_littleend(fd, format);
    write_ushort_littleend(fd, nchannels);
    write_ulong_littleend(fd, sample_rate);
    write_ulong_littleend(fd, av_bps);
    write_ushort_littleend(fd, block_align);
    write_ushort_littleend(fd, bits_per_sample);
    write(fd, "data", 4);
    write_ulong_littleend(fd, data_size);
}

void usffd_write_header(int fd, const USFFD* usffd, long* file_encoding)
{
    lseek(fd, 0, SEEK_SET);

    switch(usffd->hdr_type)
    {
      case USFFD_HDR_NONE:
	*file_encoding = usffd->data_encoding;
	break;
      case USFFD_HDR_AU:
	usffd_write_au_header(fd, usffd, file_encoding);
	break;
      case USFFD_HDR_WAV:
	usffd_write_wav_header(fd, usffd, file_encoding);
	break;
    }
}

long usffd_select_encoding(enum usffd_audio_header_t type, long encoding)
{
    switch(type)
    {
      case USFFD_HDR_NONE:
	return encoding;

      case USFFD_HDR_AU:
	switch(encoding)
	{
	  case USFFD_SIGBYTE:
	  case USFFD_UNSIGBYTE:
	    return USFFD_SIGBYTE;
	  case USFFD_G711_ULAW:
	    return USFFD_G711_ULAW;
	  case USFFD_G711_ALAW:
	    return USFFD_G711_ALAW;
	  case USFFD_SIGWORDB:
	  case USFFD_UNSIGWORDB:
	  case USFFD_SIGWORDL:
	  case USFFD_UNSIGWORDL:
	    return USFFD_SIGWORDB;
	}

      case USFFD_HDR_WAV:
	switch(encoding)
	{
	  case USFFD_SIGBYTE:
	  case USFFD_UNSIGBYTE:
	    return USFFD_UNSIGBYTE;
	  case USFFD_G711_ULAW:
	    return USFFD_G711_ULAW;
	  case USFFD_G711_ALAW:
	    return USFFD_G711_ALAW;
	  case USFFD_SIGWORDB:
	  case USFFD_UNSIGWORDB:
	  case USFFD_SIGWORDL:
	  case USFFD_UNSIGWORDL:
	    return USFFD_SIGWORDL;
	}
    }
#ifdef DEBUG
    printf("## usffd.c:usffd_select_encoding(): Unknown hdr type: %d\n",type);
#endif /* DEBUG */
    return 0;			/* dmy */
}

Bool usffd_encoding_available(enum usffd_audio_header_t type, long encoding)
{
    switch(type)
    {
      case USFFD_HDR_NONE:
	return True;
      case USFFD_HDR_AU:
	if(encoding == USFFD_G711_ULAW ||
	   encoding == USFFD_G711_ALAW ||
	   encoding == USFFD_SIGBYTE ||
	   encoding == USFFD_SIGWORDB)
	    return True;
	break;
      case USFFD_HDR_WAV:
	if(encoding == USFFD_UNSIGBYTE ||
	   encoding == USFFD_SIGWORDL ||
	   encoding == USFFD_G711_ALAW ||
	   encoding == USFFD_G711_ULAW)
	    return True;
	break;
    }
    return False;
}
