#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "datatype.h"

#define AUMAGIC 0x2e736e64	/* ".snd" in big-order */

/* Sun Audio File Encoding Tags */
#define AUDIO_FILE_ENCODING_MULAW_8     1       /* 8-bit ISDN u-law */
#define AUDIO_FILE_ENCODING_ALAW_8      27      /* 8-bit ISDN A-law */
#define AUDIO_FILE_ENCODING_LINEAR_8    2       /* 8-bit linear PCM */
#define AUDIO_FILE_ENCODING_LINEAR_16   3       /* 16-bit linear PCM */


/* audio file header
(4) magic
(4) header size
(4) data size
(4) data encoding
(4) sample rate
(4) number of channels
*/


typedef struct au_hdr
{
    U32 magic;
    U32 hdr_size;
    U32 data_size;
    U32 data_encoding;
    U32 sample_rate;
    U32 nchannels;
} hdr;


U32 read_ulong_bigend(int fd);
void write_ulong_bigend(int fd, U32 value);
U32 read_default(char* prompt, U32 default_value);
int read_header(int fd, hdr* h);
int write_header(int fd, hdr* h, char* comment);

int little_endian;

void usage(void)
{
    fprintf(stderr, "Usage: aufix [-e <encoding>] [-r <sample-rate>] [-n <nchannels>] [-c <comment>] [-h] filename\n");
    fprintf(stderr, "       encoding:\n");
    fprintf(stderr, "       %d: 8-bit ISDN u-law\n",
	    AUDIO_FILE_ENCODING_MULAW_8);
    fprintf(stderr, "       %d: 8-bit ISDN A-law\n",
	    AUDIO_FILE_ENCODING_ALAW_8);
    fprintf(stderr, "       %d: linear PCM\n",
	    AUDIO_FILE_ENCODING_LINEAR_8);
    fprintf(stderr, "       %d: 16-bit ISDN u-law\n",
	    AUDIO_FILE_ENCODING_LINEAR_16);
}

void main(int argc, char** argv)
{
    int opt_encoding    = 0;
    int opt_nchannels   = 0;
    int opt_sample_rate = 0;
    hdr h, horig;
    char* filename;
    char* comment;
    int i, fd;

    if(argc < 2)
    {
	usage();
	exit(1);
    }

    /* check byte order */
    little_endian = 1;
    if(*(char *)&little_endian != 1)
	little_endian = 0;

    /* parse args. */
    filename = comment = NULL;
    for(i = 1; i < argc; i++)
    {
	if(!strcmp(argv[i], "-e"))
	{
	    if(++i == argc) { usage(); exit(1); }
	    opt_encoding = 1;
	    h.data_encoding = atoi(argv[i]);
	}
	else if(!strcmp(argv[i], "-r"))
	{
	    if(++i == argc) { usage(); exit(1); }
	    opt_sample_rate = 1;
	    h.sample_rate = atoi(argv[i]);
	}
	else if(!strcmp(argv[i], "-n"))
	{
	    if(++i == argc) { usage(); exit(1); }
	    opt_nchannels = 1;
	    h.nchannels = atoi(argv[i]);
	}
	else if(!strcmp(argv[i], "-c"))
	{
	    if(++i == argc) { usage(); exit(1); }
	    comment = argv[i];
	}
	else if(!strcmp(argv[i], "-h"))
	{
	    usage();
	    exit(0);
	}
	else if(argv[i][0] == '-')
	{
	    usage();
	    exit(1);
	}
	else
	{
	    filename = argv[i];
	    break;
	}
    }

    if(filename == NULL)
    {
	usage();
	exit(1);
    }

    if((fd = open(filename, O_RDWR)) < 0)
    {
	perror(filename);
	exit(1);
    }
    if(read_header(fd, &horig) < 0)
    {
	perror(filename);
	close(fd);
	exit(1);
    }
    if(!opt_encoding)
	h.data_encoding = read_default("encoding", horig.data_encoding);
    if(!opt_sample_rate)
	h.sample_rate = read_default("sample rate", horig.sample_rate);
    if(!opt_nchannels)
	h.nchannels = read_default("number of channels", horig.nchannels);
    if(write_header(fd, &h, comment) < 0)
    {
	perror(filename);
	close(fd);
	exit(1);
    }

    close(fd);
    exit(0);
}

U32 read_ulong_bigend(int fd)
{
    unsigned char c[4];

    if(read(fd, c, 4) != 4)
	return 0;
    printf("%02x %02x %02x %02x\n", c[0], c[1], c[2], c[3]);
    return  ((U32)c[3])		|
	    (((U32)c[2]) << 8)	|
	    (((U32)c[1]) << 16)	|
	    (((U32)c[0]) << 24);
}

void write_ulong_bigend(int fd, U32 value)
{
    if(little_endian)
	value = ((value << 24) |
		 ((value & 0x0000ff00) << 8) |
		 ((value & 0x00ff0000) >> 8) |
		 (value >> 24));
    write(fd, &value, 4);
}

U32 read_default(char* prompt, U32 default_value)
{
    char buff[BUFSIZ];
    int n;

    sprintf(buff, "%s (default %d): ", prompt, default_value);
    write(1, buff, strlen(buff));
    n = read(0, buff, BUFSIZ);
    if(n <= 0 || (n == 1 && buff[0] == '\n'))
	return default_value;
    return (U32)atoi(buff);
}

int read_header(int fd, hdr* h)
{
    if(lseek(fd, 0, SEEK_CUR) < 0)
	return -1;
    h->magic         = read_ulong_bigend(fd);
    h->hdr_size      = read_ulong_bigend(fd);
    h->data_size     = read_ulong_bigend(fd);
    h->data_encoding = read_ulong_bigend(fd);
    h->sample_rate   = read_ulong_bigend(fd);
    h->nchannels     = read_ulong_bigend(fd);
    return 0;
}

int write_header(int fd, hdr* h, char* comment)
{
    struct stat sbuf;
    size_t clen;
    if(comment == NULL)
	comment = "";
    clen = strlen(comment);

    if(fstat(fd, &sbuf) < 0)
	return -1;

    if(lseek(fd, 0, SEEK_SET) < 0)
	return -1;

    /* magic */
    write_ulong_bigend(fd, AUMAGIC);

    /* header size */
    write_ulong_bigend(fd, 48 + clen);

    /* data size */
    write_ulong_bigend(fd, sbuf.st_size - 48 - clen);

    /* data encoding */
    write_ulong_bigend(fd, h->data_encoding);

    /* sample rate */
    write_ulong_bigend(fd, h->sample_rate);

    /* number of channels */
    write_ulong_bigend(fd, h->nchannels);

    /* comment */
    write(fd, comment, clen);

    return 0;
}
