#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include "convolution.h"
#include "timer.h"

void do_filter(const char* def_file, const char* infile, const char* outfile);
char* readtty(long* nread);

void usage(void)
{
    fputs("Usage: swfilter [-v] [-h] filterdef infile outfile\n", stderr);
}

int verbose = 0;
void main(int argc, char** argv)
{
    int arg;
    arg = 1;

    for(arg = 1; arg < argc; arg++)
    {
	if(!strcmp(argv[arg], "-v"))
	    verbose = 1;
	else if(!strcmp(argv[arg], "-h"))
	{
	    usage();
	    exit(0);
	}
	else if(argv[arg][0] == '-')
	{
	    usage();
	    exit(1);
	}
	else
	    break;
    }
    if(arg + 3  != argc)
    {
	usage();
	exit(1);
    }
    do_filter(argv[arg], argv[arg+1], argv[arg+2]);
    exit(0);
}

int write_open(const char* file)
{
    int fd;

    if(!strcmp(file, "-"))
	return 1;

    if((fd = open(file, O_WRONLY | O_CREAT | O_TRUNC)) < 0)
    {
	perror(file);
	exit(1);
    }

    return fd;
}

char* readfile(const char* fname, long* n)
{
    int fd;
    struct stat buf;
    char* text;

    if((fd = open(fname, O_RDONLY)) < 0)
    {
	perror(fname);
	exit(1);
    }

    if(fstat(fd, &buf) < 0)
    {
	perror(fname);
	close(fd);
	exit(1);
    }

    if(!S_ISREG(buf.st_mode))
    {
	fprintf(stderr, "%s: Not a regular file.\n", fname);
	close(fd);
	exit(1);
    }

    if((text = mmap(0, buf.st_size, PROT_READ,
		    MAP_SHARED, fd, 0)) == (caddr_t) - 1)
    {
	perror("mmap");
	close(fd);
	exit(1);
    }
    close(fd);
    *n = buf.st_size;
    return text;
}

void read_filter_def(const char* file, double* h)
{
    char* text;
    char* p;
    int i;
    long n;
    int map;

    if(!strcmp(file, "-"))
    {
	text = readtty(&n);
	map = 0;
    }
    else
    {
	text = readfile(file, &n);
	map = 1;
    }
    p = text - 1;
    i = 0;
    for(; i <= CONVOLUTION_SIZE/2; p = strchr(p, '\n'))
    {
	if(p == NULL)
	{
	    fprintf(stderr, "Too few lines in %s\n", file);
	    if(map)
		munmap(text, n);
	    else
		free(text);
	    exit(1);
	}
	p++;			/* skip '\n' */
	while(*p == ' ' || *p == '\t')
	    p++;
	if(*p == '#' || *p == '\n' || *p == '\0')
	    continue;
	h[i++] = atof(p);
	if(verbose)
	{
	    if(h[i] == 0) h[i] = 0;
	    if(h[i-1] < 0)
		printf("[%2d]%.3e", i-1, h[i-1]);
	    else
		printf("[%2d] %.3e", i-1, h[i-1]);
	    if(i != 1 && i != CONVOLUTION_SIZE/2 && i % 4 == 0)
		printf("\n");
	    else
		printf("\t");
	}
    }
    if(verbose)
	printf("\n");
    if(map)
	munmap(text, n);
    else
	free(text);
}

#define CONVOLUTION_STEP 8192
void do_filter(const char* def_file, const char* infile, const char* outfile)
{
    double h[CONVOLUTION_SIZE/2+1];
    short* data;
    long n, ncnv, rest_samples;
    int i;
    static short buff[CONVOLUTION_STEP];
    double t = 0;
    int ofd;
    int map;

    read_filter_def(def_file, h);
    mkconvolution(h);

    if(verbose)
    {
	printf("\n");
	for(i = 0; i <= CONVOLUTION_SIZE/2; i++)
	{
	    if(h[i] == 0) h[i] = 0;
	    if(h[i] < 0)
		printf("[%2d]%.3e", i, h[i]);
	    else
		printf("[%2d] %.3e", i, h[i]);
	    if(i != 0 && i != CONVOLUTION_SIZE/2-1 && (i+1) % 4 == 0)
		printf("\n");
	    else
		printf("\t");
	}
	printf("\n\n");
    }

    ofd = write_open(outfile);

    if(!strcmp(infile, "-"))
    {
	data = (short *)readtty(&n);
	map = 0;
    }
    else
    {
	data = (short *)readfile(infile, &n);
	map = 1;
    }

    rest_samples = n / sizeof(short);
    ncnv = 0;

    while(rest_samples > 0)
    {
	i = CONVOLUTION_STEP;
	if(i > rest_samples)
	    i = rest_samples;
	if(verbose)
	    t = get_current_calender_time();
	convolution_proc(data, n / sizeof(short), ncnv,
			 buff, i, h);
	write(ofd, buff, i * sizeof(short));
	ncnv += i;
	rest_samples -= i;

	if(verbose)
	{
	    t = get_current_calender_time() - t;
	    if(t > 0)
		printf("convolution %.1f%% (%.2fK samples/sec)\n",
		       (double)ncnv / n, 0.001 * (double)i / t);
	}
    }
    close(ofd);
    if(map)
	munmap((char *)data, n);
    else
	free(data);
}
