#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <limits.h>
#include "convolution.h"
#include "timer.h"

void doit(const char* ifile, const char* ofile);
void doit_stream(int ofd, double* hl, double* hr);
void doit_file(const char* ifile, int ofd, double* hl, double* hr);

char* readfile(const char* fname, long* n);

void usage(void)
{
    fputs("Usage: swmono2stereo [-v] infile outfile\n", stderr);
}

int verbose;
void main(int argc, char** argv)
{
    char* ifile;
    char* ofile;
    int arg;

    if(argc < 3)
    {
	usage();
	exit(1);
    }

    if(!strcmp(argv[1], "-v"))
    {
	arg = 2;
	verbose = 1;
    }
    else
    {
	arg = 1;
    }

    if(arg + 2 != argc)
    {
	usage();
	exit(1);
    }

    ifile = argv[arg++];
    ofile = argv[arg++];

    doit(ifile, ofile);
}

#if 0
void mkconvolution2(double* h)
{
    static double hx[CONVOLUTION_SIZE];
    static double hy[CONVOLUTION_SIZE];
    int i;
    double t;

    for(i = 0; i <= CONVOLUTION_SIZE/2; i++)
	hx[i] = h[i];
    for(i = 1; i < CONVOLUTION_SIZE/2; i++)
	hx[CONVOLUTION_SIZE - i] = hx[i];
    for(i = 0; i < CONVOLUTION_SIZE; i++)
	hy[i] = 0.0;

    fft(hx, hy, -CONVOLUTION_SIZE);

    for(i = 0; i < CONVOLUTION_SIZE; i++)
    {
	double x, y;
	double wx, wy;
	wx = cos(i * M_PI / CONVOLUTION_SIZE);
	wy = sin(i * M_PI / CONVOLUTION_SIZE);

	x = hx[i];
	y = hy[i];

	hx[i] = x * wx - y * wy;
    }

    for(i = 0; i < CONVOLUTION_SIZE/2; i++)
    {
	double tmp;

	tmp = hx[i];
	hx[i] = hx[i + CONVOLUTION_SIZE/2];
	hx[i + CONVOLUTION_SIZE/2] = tmp;
    }

    /* Window */
    t = -M_PI;
    for(i = 0; i <= CONVOLUTION_SIZE/2; i++)
    {
	hx[i] *= (0.50 + 0.50 * cos(t)) / CONVOLUTION_SIZE;
	t += 2.0 * M_PI / CONVOLUTION_SIZE;
    }

    bcopy(hx, h, (CONVOLUTION_SIZE/2 + 1) * sizeof(double));
}
#endif

#define CONVOLUTION_STEP 8192
void doit(const char* ifile, const char* ofile)
{
    int i;
    double hl[CONVOLUTION_SIZE/2+1], hr[CONVOLUTION_SIZE/2+1];
    int ofd;

    if(!strcmp(ofile, "-"))
	ofd = 1;
    else
	if((ofd = open(ofile, O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0)
	{
	    perror(ofile);
	    exit(1);
	}

    hl[0] = 1.0;
    hr[0] = 0.0;
    hl[1] = 1.0;
    hr[1] = 0.0;
    hl[2] = 1.0;
    hr[2] = 0.0;
    hl[3] = 1.0;
    hr[3] = 0.0;
    hl[4] = 1.0;
    hr[4] = 0.0;

    for(i = 5; i <= CONVOLUTION_SIZE/2; i++)
    {
	double y;

	y = 0.5 * (1 + cos((M_PI / M_LN2) * log((double)i)));
	hl[i] = 1 - y;
	hr[i] = y;
    }

    mkconvolution(hl);
    mkconvolution(hr);

    if(!strcmp(ifile, "-"))
	doit_stream(ofd, hl, hr);
    else
	doit_file(ifile, ofd, hl, hr);
    close(ofd);
}

void doit_file(const char* ifile, int ofd, double* hl, double* hr)
{
    short* data;
    long ncnv, n, data_size, rest_samples;
    double t = 0.0;
    static short buffl[CONVOLUTION_STEP], buffr[CONVOLUTION_STEP];
    static short bufflr[2 * CONVOLUTION_STEP];

    data = (short *)readfile(ifile, &data_size);
    n = data_size / sizeof(short) - 8;
    rest_samples = n;
    ncnv = 0;

    while(rest_samples > 0)
    {
	int i, j;

	i = CONVOLUTION_STEP;
	if(i > rest_samples)
	    i = rest_samples;
	if(verbose)
	    t = get_current_calender_time();
	convolution_proc(data + 8, n, ncnv, buffl, i, hl);
	convolution_proc(data,     n, ncnv, buffr, i, hr);

	for(j = 0; j < i; j++)
	{
	    bufflr[j * 2    ] = buffl[j];
	    bufflr[j * 2 + 1] = buffr[j];
	}

	write(ofd, bufflr, 2 * i * sizeof(short));
	ncnv += i;
	rest_samples -= i;

	if(verbose)
	{
	    t = get_current_calender_time() - t;
	    if(t > 0)
		printf("convolution %.1f%% (%.2fK samples/sec)\n",
		       (double)ncnv / n * 100, 0.001 * (double)i / t);
	}
    }
    close(ofd);
    munmap((char *)data, data_size);
}

#define READBLOCKLEN (16 * 1024)
#define CLIP_SHORT(x) ((x) < SHRT_MIN ? SHRT_MIN : (x) > SHRT_MAX ? SHRT_MAX : (short)(x))
void doit_stream(int ofd, double* hl, double* hr)
{
    static short ibuff[CONVOLUTION_SIZE + READBLOCKLEN];
    static short obufflr[2 * READBLOCKLEN];
    short outp;
    int n;

    fread(ibuff, 2, CONVOLUTION_SIZE, stdin);
    while((n = fread(ibuff + CONVOLUTION_SIZE, 2, READBLOCKLEN, stdin))
	  == READBLOCKLEN)
    {
	int i;

	outp = 0;
	for(i = 0; i < READBLOCKLEN; i++)
	{
	    double vl, vr;
	    int j, k;

	    vl = hl[0] * ibuff[i] +
		hl[CONVOLUTION_SIZE/2] * ibuff[i + CONVOLUTION_SIZE/2];
	    vr = hr[0] * ibuff[i] +
		hr[CONVOLUTION_SIZE/2] * ibuff[i + CONVOLUTION_SIZE/2];
	    for(j = 1, k = CONVOLUTION_SIZE - 1;
		j < CONVOLUTION_SIZE/2;
		j++, k--)
	    {
		vl += hl[j] * ((int)ibuff[i + j] + (int)ibuff[i + k]);
		vr += hr[j] * ((int)ibuff[i + j] + (int)ibuff[i + k]);
	    }
	    obufflr[outp++] = CLIP_SHORT(vl);
	    obufflr[outp++] = CLIP_SHORT(vr);
	}
	write(ofd, obufflr, 4 * READBLOCKLEN);
	memcpy(ibuff, ibuff + READBLOCKLEN, 2 * CONVOLUTION_SIZE);
    }

    if(n < 0)
	perror("read");
}

char* readfile(const char* fname, long* n)
{
    int fd;
    struct stat buf;
    char* text;

    if((fd = open(fname, O_RDONLY)) < 0)
    {
	perror(fname);
	exit(1);
    }

    if(fstat(fd, &buf) < 0)
    {
	perror(fname);
	close(fd);
	exit(1);
    }

    if(!S_ISREG(buf.st_mode))
    {
	fprintf(stderr, "%s: Not a regular file.\n", fname);
	close(fd);
	exit(1);
    }

    if((text = mmap(0, buf.st_size, PROT_READ,
		    MAP_SHARED, fd, 0)) == (caddr_t) - 1)
    {
	perror("mmap");
	close(fd);
	exit(1);
    }
    close(fd);
    *n = buf.st_size;
    return text;
}
