#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include "libcompress.h"
#include "timer.h"

char* readfile(const char* fname, long* n)
{
    int fd;
    struct stat buf;
    char* text;

    if((fd = open(fname, O_RDONLY)) < 0)
    {
	perror(fname);
	exit(1);
    }

    if(fstat(fd, &buf) < 0)
    {
	perror(fname);
	close(fd);
	exit(1);
    }

    if(!S_ISREG(buf.st_mode))
    {
	fprintf(stderr, "%s: Not a regular file.\n", fname);
	close(fd);
	exit(1);
    }

    if((text = mmap(0, buf.st_size, PROT_READ,
		    MAP_SHARED, fd, 0)) == (caddr_t) - 1)
    {
	perror("mmap");
	close(fd);
	exit(1);
    }
    close(fd);
    *n = buf.st_size;
    return text;
}

char* progname;
void usage(void)
{
    fprintf(stderr, "Usage: %s ", progname);
    fputs(" [-v] [-q <4|3|2>] [-d] [-h] infile outfile\n", stderr);
    fputs("       -d\tdecompress\n", stderr);
    fputs("       -q\tquantizer (2,3 or 4 bits)", stderr);
    fputs("       -v\tverbose mode\n", stderr);
    fputs("       -h\tgive this help\n", stderr);
}

void main(int argc, char** argv)
{
    char* ifile;
    char* ofile;
    unsigned char* data;
    char* obuff;
    int i, decompress_flag, verbose;
    long data_size, out_size = 0, in_size = 0;
    int ofd;
    int qsize = 4;
    double start_time = 0;

    if((progname = strrchr(argv[0], '/')) == NULL)
	progname = argv[0];
    else
	progname++;

    decompress_flag = 0;
    verbose = 0;
    for(i = 1; i < argc; i++)
    {
	if(!strcmp(argv[i], "-q"))
	{
	    if(++i == argc)
	    {
		usage();
		exit(1);
	    }
	    qsize = atoi(argv[i]);
	    if(qsize != 2 && qsize != 3 && qsize != 4)
	    {
		fprintf(stderr, "%s: Invalid quantizer\n", argv[i]);
		usage();
		exit(1);
	    }
	}
	else if(!strcmp(argv[i], "-d"))
	    decompress_flag = 1;
	else if(!strcmp(argv[i], "-v"))
	    verbose = 1;
	else if(!strcmp(argv[i], "-h"))
	{
	    usage();
	    exit(0);
	}
	else if((argv[i][0] == '-' && argv[i][1] == '\0') ||
		argv[i][0] != '-')
	    break;
	else
	{
	    usage();
	    exit(1);
	}
    }
    if(i + 2 != argc)
    {
	usage();
	exit(1);
    }
    ifile = argv[i];
    ofile = argv[i+1];

    if(!strcmp(ofile, "-"))
	ofd = 1;
    else
	if((ofd = open(ofile, O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0)
	{
	    perror(ofile);
	    exit(1);
	}

    data = (unsigned char *)readfile(ifile, &data_size);

    if(decompress_flag)
    {
	switch(qsize)
	{
	  case 4:
	    in_size = data_size;
	    out_size = data_size * 2;
	    break;
	  case 3:
	    in_size = data_size - data_size % 3;
	    out_size = in_size / 3 * 8;
	    break;
	  case 2:
	    in_size = data_size;
	    out_size = data_size * 4;
	    break;
	}
#ifdef SWCOMPRESS
	out_size *= 2;
#endif
    }
    else
    {
	long nsamples;

#ifdef SWCOMPRESS
	nsamples = data_size / 2;
#else
	nsamples = data_size;
#endif

	switch(qsize)
	{
	  case 4:
	    out_size = nsamples / 2;
	    in_size = out_size * 2;
	    break;
	  case 3:
	    in_size = nsamples - nsamples % 8;
	    out_size = in_size / 8 * 3;
	    break;
	  case 2:
	    out_size = nsamples / 4;
	    in_size = out_size * 4;
	    break;
	}
    }

    if((obuff = (char *)malloc(out_size)) == NULL)
    {
	perror("malloc");
	exit(1);
    }

    if(verbose)
	start_time = get_current_calender_time();

#ifdef SWCOMPRESS
#define CAST (short *)
#else
#define CAST
#endif

    if(decompress_flag)
    {
	switch(qsize)
	{
	  case 4: do_uldecompress4(0, data, in_size, CAST obuff); break;
	  case 3: do_uldecompress3(0, data, in_size, CAST obuff); break;
	  case 2: do_uldecompress2(0, data, in_size, CAST obuff); break;
	}
    }
    else
    {
	switch(qsize)
	{
	  case 4: do_ulcompress4(0, CAST data, in_size, obuff); break;
	  case 3: do_ulcompress3(0, CAST data, in_size, obuff); break;
	  case 2: do_ulcompress2(0, CAST data, in_size, obuff); break;
	}
    }

    if(verbose)
	printf("compress: %.2fK bytes/sec\n",
	       in_size / (get_current_calender_time() - start_time) / 1000.0);
    write(ofd, obuff, out_size);

    munmap(data, data_size);
    exit(0);
}
